from typing import Dict

import torch
from skoots.validate.stats import get_volume, get_surface_area
from torch import Tensor


def stats_per_instance(x: Tensor, anisotropy: Tensor) -> Dict[str, Tensor]:
    """

    :param x:
    :param anisotropy:
    :return:
    """
    summary = {}

    object_id = torch.unique(x)
    object_id = object_id[object_id > 0]
    summary["id"] = object_id

    summary["volume"] = torch.empty_like(object_id)
    summary["surface_area"] = torch.empty_like(object_id)

    for i, id in enumerate(object_id):
        summary["volume"][i] = get_volume(x.eq(id), anisotropy)
        summary["surface_area"][i] = get_surface_area(x.eq(id), anisotropy)

    return summary


def compare(ground_truth, predictions) -> Dict[str, Tensor]:
    raise NotImplementedError



import argparse
import os.path

import torch
import torch.multiprocessing as mp
import torch.nn as nn

from skoots.config import get_cfg_defaults
from skoots.lib.mp_utils import find_free_port
from skoots.lib.utils import cfg_to_bism_model
from skoots.train.engine import train
import warnings
import logging

torch.set_float32_matmul_precision("high")


def load_cfg_from_file(args: argparse.Namespace):
    """Load configurations."""
    # Set configurations
    cfg = get_cfg_defaults()
    if os.path.exists(args.config_file):
        cfg.merge_from_file(args.config_file)
    else:
        raise ValueError("Could not find config file from path!")
    cfg.freeze()

    return cfg


def main():
    parser = argparse.ArgumentParser(description="SKOOTS Training Parameters")
    parser.add_argument("--config-file", type=str, help="YAML config file for training")
    parser.add_argument(
        "--log",
        type=int,
        default=3,
        help="Log Level: 0-Debug, 1-Info, 2-Warning, 3-Error, 4-Critical",
    )

    args = parser.parse_args()

    # Set logging level
    _log_map = [
        logging.DEBUG,
        logging.INFO,
        logging.WARNING,
        logging.ERROR,
        logging.CRITICAL,
    ]

    logging.basicConfig(
        level=_log_map[args.log],
        format="[%(asctime)s] skoots-train [%(levelname)s]: %(message)s",
        force=True,
    )

    cfg = load_cfg_from_file(args)
    model: nn.Module = cfg_to_bism_model(cfg)  # This is our skoots torch model

    if cfg.TRAIN.PRETRAINED_MODEL_PATH and os.path.exists(cfg.TRAIN.PRETRAINED_MODEL_PATH[0]):
        checkpoint = torch.load(cfg.TRAIN.PRETRAINED_MODEL_PATH[0])
        state_dict = (
            checkpoint
            if not "model_state_dict" in checkpoint
            else checkpoint["model_state_dict"]
        )
        model.load_state_dict(state_dict)
    else:
        warnings.warn(f'Could not find file at path: {cfg.TRAIN.PRETRAINED_MODEL_PATH[0]}. Model has not been loaded.')

    port = find_free_port()
    world_size = cfg.SYSTEM.NUM_GPUS if cfg.TRAIN.DISTRIBUTED else 1

    mp.spawn(train, args=(port, world_size, model, cfg, args.log), nprocs=world_size, join=True)


if __name__ == "__main__":
    import sys

    sys.exit(main())



from torch import Tensor
from typing import Tuple, List

from torch import Tensor


def get_adjacent_labels(x: Tensor, y: Tensor) -> List[Tuple[int, int]]:
    """
    calculates which masks of a signle object have two labels (due to the border)

    :param x:
    :param y:
    :return:
    """

    # Could ideally use the cantor pairing function but might overflow???
    # this seems to be safe and memory efficient
    z0 = (x + y).unique().tolist()
    z1 = (x * y).unique().tolist()

    identical = []

    for _x in x.unique():
        for _y in y.unique():
            if _x == 0 or _y == 0:
                continue

            if (_x + _y) in z0 and (_x * _y) in z1:
                identical.append((_x.item(), _y.item()))
    # identical = identical if len(identical) > 0 else None
    return identical



from typing import List
from typing import Tuple

import torch
import torch.nn.functional as F
from torch import Tensor


@torch.jit.script
def _compute_zero_padding(kernel_size: List[int]) -> Tuple[int, int, int]:
    r"""Utility function that computes zero padding tuple.
    Adapted from Kornia
    """
    computed: List[int] = [(k - 1) // 2 for k in kernel_size]
    return computed[0], computed[1], computed[2]


@torch.jit.script
def _get_binary_kernel3d(window_size: int, device: str) -> Tensor:
    r"""Creates a symmetric binary kernel to extract the patches. If the window size
    is HxWxD will create a (H*W)xHxW kernel.

    Adapted from a 2D Kornia implementation

    """
    window_range: int = int(window_size**3)
    kernel: Tensor = torch.zeros(
        (window_range, window_range, window_range), device=device
    )
    for i in range(window_range):
        kernel[i, i, i] += 1.0
    kernel = kernel.view(-1, 1, window_size, window_size, window_size)

    # get rid of all zero kernels
    ind = torch.nonzero(kernel.view(kernel.shape[0], -1).sum(1))
    return kernel[ind[:, 0], ...]


@torch.jit.script
def _get_binary_kernel2d(window_size: int, device: str) -> Tensor:
    r"""Creates a symmetric binary kernel to extract the patches. If the window size
    is HxWxD will create a (H*W)xHxW kernel.

    Adapted from a 2D Kornia implementation

    """
    window_range: int = int(window_size**3)
    kernel: Tensor = torch.zeros((window_range, window_range, 1), device=device)
    for i in range(window_range):
        kernel[i, i, 0] += 1.0
    kernel = kernel.view(-1, 1, window_size, window_size, 1)

    # get rid of all zero kernels
    ind = torch.nonzero(kernel.view(kernel.shape[0], -1).sum(1))
    return kernel[ind[:, 0], ...]


# re-implemented from torchvision.tensor.functional
def _get_gaussian_kernel1d(kernel_size: int, sigma: float) -> Tensor:
    ksize_half = (kernel_size - 1) * 0.5

    x = torch.linspace(-ksize_half, ksize_half, steps=kernel_size)
    pdf = torch.exp(-0.5 * (x / sigma).pow(2))
    kernel1d = pdf / pdf.sum()

    return kernel1d


# re-implemented from torchvision.tensor.functional
def _get_gaussian_kernel2d(
    kernel_size: List[int], sigma: List[float], dtype: torch.dtype, device: torch.device
) -> Tensor:
    kernel1d_x = _get_gaussian_kernel1d(kernel_size[0], sigma[0]).to(
        device, dtype=dtype
    )
    kernel1d_y = _get_gaussian_kernel1d(kernel_size[1], sigma[1]).to(
        device, dtype=dtype
    )
    kernel2d = torch.mm(kernel1d_y[:, None], kernel1d_x[None, :])
    return kernel2d


# expanded to 3D
def _get_gaussian_kernel3d(
    kernel_size: List[int], sigma: List[float], dtype: torch.dtype, device: torch.device
) -> Tensor:
    kernel1d_x = _get_gaussian_kernel1d(kernel_size[0], sigma[0]).to(
        device, dtype=dtype
    )
    kernel1d_y = _get_gaussian_kernel1d(kernel_size[1], sigma[1]).to(
        device, dtype=dtype
    )
    kernel1d_z = _get_gaussian_kernel1d(kernel_size[2], sigma[2]).to(
        device, dtype=dtype
    )

    kernel3d = (kernel1d_x[:, None] @ kernel1d_y[None, :]).unsqueeze(-1) @ kernel1d_z[
        None, :
    ]
    return kernel3d


@torch.jit.script
def gauss_filter(input: Tensor, kernel: List[int], sigma: List[float]) -> Tensor:
    """
    gaussian filter of a 3D tensor

    :param input: (B, C, X, Y, Z)
    :param kernel: [int, int, int]
    :param sigma: [float, float, float]
    :return: blured image
    """
    padding: Tuple[int, int, int] = _compute_zero_padding(kernel)
    kernel: Tensor = _get_gaussian_kernel3d(kernel, sigma, input.dtype, input.device)
    kernel = kernel.expand(
        input.shape[1], 1, kernel.shape[0], kernel.shape[1], kernel.shape[2]
    )

    features: Tensor = F.conv3d(
        input, kernel, padding=padding, stride=(1, 1, 1), groups=input.shape[1]
    )

    return features


@torch.jit.script
def binary_erosion(image: Tensor) -> Tensor:
    """
    Performs binary erosion on a 5D Tensor.

    Shapes:
        - input: :math:`(B, C, X, Y, Z)`
        - output: :math:`(C, C, X, Y, Z)`


    :param image: binary image
    :return: eroded image
    """
    device = str(image.device)
    kernel = _get_binary_kernel3d(3, device)
    padding = _compute_zero_padding((3, 3, 3))

    b, c, h, w, d = image.shape
    # map the local window to single vector
    features: Tensor = F.conv3d(
        image.reshape(b * c, 1, h, w, d), kernel, padding=padding, stride=1
    )
    return features.min(dim=1)[0].unsqueeze(0)


@torch.jit.script
def binary_dilation(image: Tensor) -> Tensor:
    """
    Performs binary dilation on a 5D Tensor.

    Shapes:
        - input: :math:`(B, C, X, Y, Z)`
        - output: :math:`(C, C, X, Y, Z)`

    :param image: binary image
    :return: dilated image
    """
    padding: Tuple[int, int, int] = _compute_zero_padding((3, 3, 3))
    kernel: Tensor = _get_binary_kernel3d(3, str(image.device))

    b, c, h, w, d = image.shape
    # map the local window to single vector
    features = F.conv3d(
        image.reshape(b * c, 1, h, w, d), kernel, padding=padding, stride=1
    )
    return torch.max(features.view(b, c, -1, h, w, d), dim=2)[0]


@torch.jit.ignore
def binary_dilation_2d(image: Tensor) -> Tensor:
    """
    Performs binary dilation on a 5D Tensor.

    Shapes:
        - input: :math:`(B, C, X, Y, Z)`
        - output: :math:`(C, C, X, Y, Z)`

    :param image: binary image
    :return: dilated image
    """
    padding: Tuple[int, int, int] = _compute_zero_padding((3, 3, 1))
    kernel: Tensor = _get_binary_kernel2d(3, str(image.device))

    b, c, h, w, d = image.shape
    # map the local window to single vector
    features = F.conv3d(
        image.reshape(b * c, 1, h, w, d), kernel, padding=padding, stride=1
    )

    return torch.max(features.view(b, c, -1, h, w, d), dim=2)[0]


def median_filter(input: Tensor) -> Tensor:
    padding: Tuple[int, int, int] = _compute_zero_padding((3, 3, 3))
    kernel: Tensor = _get_binary_kernel3d(3, input.dtype, input.device)
    b, c, h, w, d = input.shape
    # map the local window to single vector
    features: Tensor = F.conv3d(
        input.reshape(b * c, 1, h, w, d), kernel, padding=padding, stride=1
    )
    return torch.median(features.view(b, c, -1, h, w, d), dim=2)[0]


def mean_filter(input: Tensor) -> Tensor:
    padding: Tuple[int, int, int] = _compute_zero_padding((3, 3, 3))
    kernel: Tensor = _get_binary_kernel3d(3, input.dtype, input.device)
    b, c, h, w, d = input.shape
    # map the local window to single vector
    features: Tensor = F.conv3d(
        input.reshape(b * c, 1, h, w, d), kernel, padding=padding, stride=1
    )
    return torch.mean(features.view(b, c, -1, h, w, d), dim=2)[0]


def dilate(input: Tensor) -> Tensor:
    padding: Tuple[int, int, int] = _compute_zero_padding((3, 3, 3))
    kernel: Tensor = _get_binary_kernel3d(3, input.dtype, input.device)
    b, c, h, w, d = input.shape
    # map the local window to single vector
    features = F.conv3d(
        input.reshape(b * c, 1, h, w, d), kernel, padding=padding, stride=1
    )
    return torch.max(features.view(b, c, -1, h, w, d), dim=2)[0]



from typing import Dict, Tuple, Union, List

import torch
import torch.nn.functional as F
from torch import Tensor


@torch.jit.script
def binary_erosion(image: Tensor) -> Tensor:
    kernel = _get_binary_kernel3d(3, image.device)
    padding = _compute_zero_padding((3, 3, 3))

    b, c, h, w, d = image.shape
    # map the local window to single vector
    features: Tensor = F.conv3d(
        image.reshape(b * c, 1, h, w, d), kernel, padding=padding, stride=1
    )
    return features.min(dim=1)[0].unsqueeze(0)


class erosion:
    def __init__(
        self, kernel_targets: int = 3, rate: float = 0.5, device: str = "cpu"
    ) -> None:
        super(erosion, self).__init__()
        self.device = device
        self.rate = rate
        self.kernel_targets = kernel_targets

        if kernel_targets % 2 != 1:
            raise ValueError("Expected Kernel target to be Odd")

        self.padding: Tuple[int, int, int] = self._compute_zero_padding(
            (kernel_targets, kernel_targets, kernel_targets)
        )
        self.kernel: torch.Tensor = self._get_binary_kernel3d(kernel_targets, device)

    def __call__(
        self, input: Union[Dict[str, torch.Tensor], torch.Tensor]
    ) -> Union[Dict[str, torch.Tensor], torch.Tensor]:
        """
        Randomly applies a gaussian blur

        :param data_dict Dict[str, torch.Tensor]: data_dictionary from a dataloader. Has keys:
            key : val
            'image' : torch.Tensor of size [C, X, Y, Z] where C is the number of colors, X,Y,Z are the mask height,
                      width, and depth
            'masks' : torch.Tensor of size [I, X, Y, Z] where I is the number of identifiable objects in the mask
            'centroids' : torch.Tensor of size [I, 3] where dimension two is the [X, Y, Z] position of the centroid
                          for instance i

        :return: data_dict Dict[str, torch.Tensor]: dictionary with identical keys as input, but with transformed values
        """
        if isinstance(input, Dict):
            self.check_inputs(input)
            return self._is_dict(input)
        elif isinstance(input, torch.Tensor):
            return self._is_tensor(input)

    def _is_dict(self, data_dict: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        if data_dict["masks"].dtype != self.kernel.dtype:
            raise ValueError(
                f'Expected Image dtype to be {self.kernel.dtype} not {data_dict["masks"].dtype}'
            )
        # We expect the image to have ndim==4 but need a batch size of 1 for median filter

        device = data_dict["image"].device
        if self.kernel.device != device:
            self.kernel = self.kernel.to(device)

        if torch.rand(1) < self.rate:
            im = data_dict["masks"].unsqueeze(0)
            b, c, h, w, d = im.shape
            # map the local window to single vector
            features: torch.Tensor = F.conv3d(
                im.reshape(b * c, 1, h, w, d),
                self.kernel,
                padding=self.padding,
                stride=1,
            )
            data_dict["masks"] = features.min(dim=1)[0]
        return data_dict

    def _is_tensor(self, input: torch.Tensor) -> torch.Tensor:
        "Assumes [C, X, Y, Z]"
        if input.device != self.kernel.device:
            raise ValueError(
                f"Expected Image Device to be {self.kernel.device} not {input.device}"
            )
        if input.dtype != self.kernel.dtype:
            raise ValueError(
                f"Expected Image dtype to be {self.kernel.dtype} not {input.dtype}"
            )
        if input.ndim != 4:
            raise ValueError(
                f"Expected Image ndim to be 4 not {input.ndim}, with shape {input.shape}"
            )

        b, c, h, w, d = input.unsqueeze(0).shape
        # map the local window to single vector
        features: torch.Tensor = F.conv3d(
            input.reshape(b * c, 1, h, w, d),
            self.kernel,
            padding=self.padding,
            stride=1,
        )
        return features.min(dim=1)[0]


def _compute_zero_padding(kernel_size: Tuple[int, int, int]) -> Tuple[int, int, int]:
    r"""Utility function that computes zero padding tuple.
    Adapted from Kornia
    """
    computed: List[int] = [(k - 1) // 2 for k in kernel_size]
    return computed[0], computed[1], computed[2]


def _get_binary_kernel3d(window_size: int, device: torch.device) -> torch.Tensor:
    r"""Creates a symetric binary kernel to extract the patches. If the window size
    is HxWxD will create a (H*W)xHxW kernel.

    Adapted from Kornia

    """
    window_range: int = int(window_size**3)
    kernel: torch.Tensor = torch.zeros(
        (window_range, window_range, window_range), device=device
    )
    for i in range(window_range):
        kernel[i, i, i] += 1.0
    kernel = kernel.view(-1, 1, window_size, window_size, window_size)

    # get rid of all zero kernels
    ind = torch.nonzero(kernel.view(kernel.shape[0], -1).sum(1))
    return kernel[ind[:, 0], ...]



from typing import Dict, TypedDict
from torch import Tensor


class DataDict(TypedDict):
    image: Tensor
    masks: Tensor
    skeletons: Dict[int, Tensor]
    baked_skeletons: Tensor
    skele_masks: Tensor




from typing import Tuple, Dict, List

import numpy as np
import torch
from numba import njit, prange
from scipy.ndimage import label
from torch import Tensor
from tqdm import tqdm

from skoots.lib.cropper import crops, get_total_num_crops


def efficient_flood_fill(skeleton: Tensor) -> Tensor:
    """
    Efficiently floods a binary skeleton mask in place by first flood filling small regions,
    then merging connected components later. Avoids memory copies when possible. Returns a skeleton mask where
    each connected component has a unique label, however these labels may not be sequential.
    I.e. unique(skeleton) -> [4, 16, 23, 24, 96]

    :param skeleton: binary skeleton mask to flood fill
    :return: Flood filled tensor
    """
    skeleton = skeleton.unsqueeze(0) if skeleton.ndim == 3 else skeleton
    total = get_total_num_crops(
        skeleton.shape, crop_size=[1000, 1000, 100], overlap=[0, 0, 0]
    )
    iterator = tqdm(
        crops(skeleton, crop_size=[1000, 1000, 200]), desc="Flood-filling small crops: "
    )
    max_id = 1
    skeletons_dict = {}

    seams_x = []  # this will be all the seams of the crops we need to check later.
    seams_y = []
    seams_z = []

    # Iterate over crops of the skeleton to perform a flood fill
    for crop, (x, y, z) in iterator:
        seams_x = seams_x + [x] if x not in seams_x else seams_x
        seams_y = seams_y + [y] if y not in seams_y else seams_y
        seams_z = seams_z + [z] if z not in seams_z else seams_z

        # crop should only contain ones and zeros and be an int. Not guaranteed by skoots.lib.cropper.crops
        crop = crop.squeeze().gt(0).int()

        crop, max_id = flood_all(
            crop, max_id + 1
        )  # max_id is the previous max - update by 1!
        w, h, d = crop.shape
        skeleton[0, x : x + w, y : y + h, z : z + d] = crop

    # We now check each crop seam for double labeled instances. Here, a collision is the same object having label (x)
    # in one crop, then label (y) in another.  We must check in all dims, x, y, and z.

    # print(f'[      ] Detecting collisions...', end='')
    # X
    collisions: List[Tuple[int, int]] = []
    for x in tqdm(seams_x, desc="Checking for duplicate IDs in X", total=len(seams_x)):
        if x > 0:
            slice_0 = skeleton[0, x, :, :]
            slice_1 = skeleton[0, x - 1, :, :]
            collisions.extend(get_adjacent_labels(slice_0, slice_1))

    # Y
    for y in tqdm(seams_y, desc="Checking for duplicate IDs in Y", total=len(seams_y)):
        if y > 0:
            slice_0 = skeleton[0, :, y, :]
            slice_1 = skeleton[0, :, y - 1, :]
            collisions.extend(get_adjacent_labels(slice_0, slice_1))

    # Z
    for z in tqdm(seams_z, desc="Checking for duplicate IDs in Z", total=len(seams_z)):
        if z > 0:
            slice_0 = skeleton[0, :, :, z]
            slice_1 = skeleton[0, :, :, z - 1]
            collisions.extend(get_adjacent_labels(slice_0, slice_1))

    # print("\r[\x1b[1;32;40m DONE \x1b[0m] Detecting collisions...")

    # Multiple collisions for each id value may exist, so we construct a graph of id values
    print(f"[      ] Constructing collision graph...", end="")
    graph: Dict[int, List[int]] = {}
    for a, b in collisions:
        if a not in graph:
            graph[a] = [b]
        else:
            graph[a].append(b)
        if b not in graph:
            graph[b] = [a]
        else:
            graph[b].append(a)
    print("\r[\x1b[1;32;40m DONE \x1b[0m] Constructing collision graph...")

    # Each skeleton, of multiple potential id values, forms a connected component in the graph
    print(f"[      ] Identifying connected components...", end="")
    cc: List[List[int]] = connected_components(graph)
    print("\r[\x1b[1;32;40m DONE \x1b[0m] Identifying connected components...")

    # We need to decide which node of the graph, (id value of a skeleton) will represent the rest of the skeleton
    # For simplicity we just choose the last value.
    to_replace: List[int] = []
    replace_with: List[int] = []
    for component in cc:
        # we always replace with the LAST value of a bunch of connected components
        a = component.pop(-1)
        replace_with.extend([a for _ in component])
        to_replace.extend(component)

    # for every pixel at position j, if it is identical to a value at position i in two_replace,
    # we replace pixel j with the new value: replace_with[i]

    collisions = [
        (a, b) for a, b in zip(to_replace, replace_with)
    ]  # replace needs a set of collisions

    print(f"[      ] Performing in place replacement of collisions...", end="")
    skeleton = replace(skeleton, collisions)  # in place replace
    print(
        "\r[\x1b[1;32;40m DONE \x1b[0m] Performing in place replacement of collisions..."
    )

    return skeleton.squeeze(0)


def flood_all(x: Tensor, id: int) -> Tuple[Tensor, int, Dict[int, Tensor]]:
    """
    Finds all features (connected components in an ndarray) and gives it a unique label from 1 to N for N components

    :param x: torch.bool tensor
    :param id: previous max id value
    :return:
    """
    x = x.gt(0).int()

    mask, max_id = label(input=x.cpu().numpy())
    mask = torch.from_numpy(mask).to(torch.int16)

    mask = mask + x.mul(id)

    return mask, mask.max()


def dfs(
    connected: List[int],
    node: int,
    graph: Dict[int, List[int]],
    visited: Dict[int, bool],
) -> List[int]:
    """depth first search for finding connected components of a graph"""
    visited[node] = True
    connected.append(node)
    for n in graph[node]:
        if not visited[n]:
            connected = dfs(connected, n, graph, visited)
    return connected


def connected_components(graph: Dict[int, List[int]]) -> List[List[int]]:
    """
    Finds all connected components in a graph of id values by performing depth
    first search.

    :param graph: input graph where each key is a node, and each value is a list of edges
    :return: list of all connected notes
    """
    cc: List[List[int]] = []

    nodes = graph.keys()
    visited = {n: False for n in graph.keys()}
    for n in nodes:
        if not visited[n]:
            connected = []
            cc.append(dfs(connected, n, graph, visited))
    return cc


@njit(parallel=True)
def _in_place_replace(
    x: np.ndarray, to_replace: np.ndarray, replace_with: np.ndarray
) -> None:
    """
    Performs an in place replacement of values in tensor x.

    Checks each location in x for a value in to_replace. if a value is in to_replace, the value is
    swapped with the associated value in replace_with.

    :param x: input nd.array
    :param to_replace: array of values to be replaced in x
    :param replace_with: array of values by which should be replaced
    :return:
    """
    assert (
        to_replace.shape == replace_with.shape
    ), "to_replace must be the same size as replace with"
    assert x.ndim == 1, "input tensor must be reveled"

    for i in prange(x.shape[0]):
        # ind = to_replace == x[i]
        # if np.any(ind):
        for j, v in enumerate(to_replace):
            if x[i] == v:
                x[i] = replace_with[j]  # replace_with[ind][0]
                break  # presumably most of the time we'll hit the actual value much sooner...


def replace(x: Tensor, collisions: List[Tuple[int, int]]) -> Tensor:
    """
    Performs an in place replacement of values in the input tensor :math:`x`.
    This function calls a just-in-time compiled numba kernel which parallelizes the replacement.

    Roughly performs this algorith:
        for i in range(x):
            for (to_replace, replace_with) in collisions:
                if x[i] == to_replace:
                    x[i] = replace_with
        return x


    :param x: Input torch.Tensor of any size
    :param collisions: List of collisions where c[0] is the value to replace, and c[1] is the value to replace with
    :return: Original tensor with modified memory
    """

    assert x.dtype == torch.int16, f"Input tensor datatype must be int16 not {x.dtype}"

    shape = x.shape
    x = x.flatten().numpy()

    to_replace = np.array([a[0] for a in collisions], dtype=np.int16).flatten()
    replace_with = np.array([a[1] for a in collisions], dtype=np.int16).flatten()

    _in_place_replace(x, to_replace=to_replace, replace_with=replace_with)

    return torch.from_numpy(x).view(shape)


def get_adjacent_labels(x: Tensor, y: Tensor) -> List[Tuple[int, int]]:
    """
    calculates which masks of a signle object have two labels (due to the border)

    :param x:
    :param y:
    :return:
    """

    # Could ideally use the cantor pairing function but might overflow???
    # this seems to be safe and memory efficient
    z0 = (x + y).unique().tolist()
    z1 = (x * y).unique().tolist()

    identical = []

    for _x in x.unique():
        for _y in y.unique():
            if _x == 0 or _y == 0:
                continue

            if (_x + _y) in z0 and (_x * _y) in z1:
                identical.append((_x.item(), _y.item()))
    # identical = identical if len(identical) > 0 else None
    return identical


if __name__ == "__main__":
    graph = {
        1: [2, 3],
        2: [1],
        3: [1, 5, 4],
        4: [5],
        5: [3],
        6: [7],
        7: [6, 8, 9],
        8: [7],
        9: [7],
    }
    cc = connected_components(graph)
    print(cc)



import math
import random
from copy import deepcopy
from typing import Dict, Tuple, Union, List, Optional, Callable

import torch
import torch.nn.functional as F
import torch.nn as nn
import torchvision.transforms.functional as ttf
from skoots.lib.morphology import binary_erosion
from skoots.lib.skeleton import bake_skeleton, skeleton_to_mask

from skoots.lib.custom_types import DataDict
from torch import Tensor
from yacs.config import CfgNode

import logging

def elastic_deform(
        image: Tensor,
        mask: Tensor,
        skeleton: Dict[int, Tensor],
        displacement_shape: Tuple[int, int, int] = (6, 6, 2),
        max_displacement: float = 0.2,
) -> Tuple[Tensor, Tensor, Dict[int, Tensor]]:
    """
    Randomly creates a deformation grid and applies to image, mask, and skeletons

    :param image: Tensor
    :param mask:  Tensor
    :param skeleton:  Dict[int, Tensor[3, N]]
    :param displacement_shape: Tuple of len 3
    :param max_displacement: float between 0 and 1
    :return:
    """

    assert image.shape == mask.shape, "image and mask shape must be the same"
    assert image.ndim == 5, f"image must be in shape: [B, C, X, Y, Z], not {image.shape}"
    assert image.device == mask.device
    assert (
            len(displacement_shape) == 3
    ), "displacement_shape must be a tuple of integers with len == 3"
    assert max_displacement < 1.0, "max displacement must not exceed 1.0"

    # for k, v in skeleton.items():
    #     assert v.shape[1] == 3, f'skeleton shape wrong: {v.shape}'

    device = image.device

    b, c, x, y, z = image.shape

    # offset are the random directon vectors
    offset: Tensor = (
        F.interpolate(
            torch.rand((1, 3, 2, 2, 2), device=device), (x, y, z), mode="trilinear"
        ).permute((0, 2, 3, 4, 1)).mul(
            torch.tensor((0.05, 0.2, 0.2), device=device).view(1, 1, 1, 1, 3)
        )

    )

    # base_grid is the identity grid. Applying base_grid alone will result in an identical image as input
    base_grid: Tensor = torch.stack(
        (
            torch.linspace(-1, 1, z, dtype=torch.float, device=device)
            .unsqueeze(0)
            .unsqueeze(0)
            .repeat(x, y, 1),
            torch.linspace(-1, 1, y, dtype=torch.float, device=device)
            .unsqueeze(0)
            .unsqueeze(-1)
            .repeat(x, 1, z),
            torch.linspace(-1, 1, x, dtype=torch.float, device=device)
            .unsqueeze(-1)
            .unsqueeze(-1)
            .repeat(1, y, z),
        ),
        dim=-1,
    ).unsqueeze(0)

    grid = (base_grid + offset).float()

    io.imsave('/home/chris/Desktop/grid.tif', grid.squeeze().permute(2, 0, 1, 3).cpu().numpy())

    # apply the deformation
    image = F.grid_sample(image.float(), grid.float(), align_corners=True, mode='nearest')
    mask = F.grid_sample(mask.float(), grid.float(), align_corners=True, mode='nearest')

    # remap grid from -1->1 to 0->shape
    # grid = (base_grid - offset).float()
    grid = grid.add(1).div(2).mul(
        torch.tensor((z, y, x), device=device).view(1, 1, 1, 1, 3)
    )
    print(f'{grid[0,34,54,22,:]=}')

    for k, skel in skeleton.items():
        sx: Tensor = skel[:, 0].long()
        sy: Tensor = skel[:, 1].long()
        sz: Tensor = skel[:, 2].long()

        ind_x = torch.logical_and(sx >= 0, sx < x)
        ind_y = torch.logical_and(sy >= 0, sy < y)
        ind_z = torch.logical_and(sz >= 0, sz < z)

        # Equivalent to a 3 way logical and
        ind = (ind_x.float() + ind_y.float() + ind_z.float()) == 3

        # grid last dim is Z, Y, X for some reason
        print(skel)
        skel[ind, :] = grid[0, sx[ind], sy[ind], sz[ind], :][:, [2, 1, 0]]
        print(skel)

        skeleton[k] = skel

    return image, mask, skeleton


@torch.jit.script
def _get_box(mask: Tensor, device: str, threshold: int) -> Tuple[Tensor, Tensor]:
    # mask in shape of 300, 400, 1 [H, W, z=1]
    nonzero = torch.nonzero(mask)  # Q, 3=[x,y,z]
    label = mask.max()

    box = torch.tensor([-1, -1, -1, -1], dtype=torch.long, device=device)

    # Recall, image in shape of [C, H, W]

    if nonzero.numel() > threshold:
        x0 = torch.min(nonzero[:, 1])
        x1 = torch.max(nonzero[:, 1])
        y0 = torch.min(nonzero[:, 0])
        y1 = torch.max(nonzero[:, 0])

        if (x1 - x0 > 0) and (y1 - y0 > 0):
            box[0] = x0
            box[1] = y0
            box[2] = x1
            box[3] = y1

    return label, box


# @torch.jit.script
def _get_affine_matrix(
        center: List[float],
        angle: float,
        translate: List[float],
        scale: float,
        shear: List[float],
        device: str,
) -> Tensor:
    # We need compute the affine transformation matrix: M = T * C * RSS * C^-1

    rot = math.radians(angle)
    sx, sy = [math.radians(s) for s in shear]

    T: Tensor = torch.eye(3, device=device)
    T[0, -1] = translate[0]
    T[1, -1] = translate[1]

    C: Tensor = torch.eye(3, device=device)
    C[0, -1] = center[0]
    C[1, -1] = center[1]

    # RSS without scaling
    # a = math.cos(rot - sy) / math.cos(sy)
    # b = -1 * math.cos(rot - sy) * math.tan(sx) / math.cos(sy) - math.sin(rot)
    # c = math.sin(rot - sy) / math.cos(sy)
    # d = math.sin(rot - sy) * math.tan(sx) / math.cos(sy) + math.cos(rot)
    a = math.cos(rot - sy) / math.cos(sy)
    b = -math.cos(rot - sy) * math.tan(sx) / math.cos(sy) - math.sin(rot)
    c = math.sin(rot - sy) / math.cos(sy)
    d = -math.sin(rot - sy) * math.tan(sx) / math.cos(sy) + math.cos(
        rot
    )  # # rotated scale shear

    RSS = torch.tensor([[a, b, 0.0], [c, d, 0.0], [0.0, 0.0, 1.0]], device=device)
    RSS = RSS * scale
    # print(RSS)

    RSS[-1, -1] = 1

    # cx, cy = center
    # tx, ty = translate
    #
    # matrix = [a, b, 0.0, c, d, 0.0]
    # matrix = [x * scale for x in matrix]
    # # Apply inverse of center translation: RSS * C^-1
    # matrix[2] += matrix[0] * (-cx) + matrix[1] * (-cy)
    # matrix[5] += matrix[3] * (-cx) + matrix[4] * (-cy)
    # # Apply translation and center : T * C * RSS * C^-1
    # matrix[2] += cx + tx
    # matrix[5] += cy + ty
    #
    #
    # # matrix = [d, -b, 0.0, -c, a, 0.0]
    # # matrix = [x / scale for x in matrix]
    # # # Apply inverse of translation and of center translation: RSS^-1 * C^-1 * T^-1
    # # matrix[2] += matrix[0] * (-cx - tx) + matrix[1] * (-cy - ty)
    # # matrix[5] += matrix[3] * (-cx - tx) + matrix[4] * (-cy - ty)
    # # # Apply center translation: C * RSS^-1 * C^-1 * T^-1
    # # matrix[2] += cx
    # # matrix[5] += cy
    #
    # matrix += [0, 0, 1]
    # matrix = torch.tensor(matrix, device=device).view(3, 3)
    # return matrix
    #
    #
    #
    return T @ C @ RSS @ torch.inverse(C)
    # return C @ torch.inverse(C) @ torch.inverse(T) @ torch.inverse(RSS)


def _get_inverse_affine_matrix(
        center: List[float],
        angle: float,
        translate: List[float],
        scale: float,
        shear: List[float],
        inverted: bool = True,
) -> List[float]:
    # Helper method to compute inverse matrix for affine transformation

    # Pillow requires inverse affine transformation matrix:
    # Affine matrix is : M = T * C * RotateScaleShear * C^-1
    #
    # where T is translation matrix: [1, 0, tx | 0, 1, ty | 0, 0, 1]
    #       C is translation matrix to keep center: [1, 0, cx | 0, 1, cy | 0, 0, 1]
    #       RotateScaleShear is rotation with scale and shear matrix
    #
    #       RotateScaleShear(a, s, (sx, sy)) =
    #       = R(a) * S(s) * SHy(sy) * SHx(sx)
    #       = [ s*cos(a - sy)/cos(sy), s*(-cos(a - sy)*tan(sx)/cos(sy) - sin(a)), 0 ]
    #         [ s*sin(a - sy)/cos(sy), s*(-sin(a - sy)*tan(sx)/cos(sy) + cos(a)), 0 ]
    #         [ 0                    , 0                                        , 1 ]
    # where R is a rotation matrix, S is a scaling matrix, and SHx and SHy are the shears:
    # SHx(s) = [1, -tan(s)] and SHy(s) = [1      , 0]
    #          [0, 1      ]              [-tan(s), 1]
    #
    # Thus, the inverse is M^-1 = C * RotateScaleShear^-1 * C^-1 * T^-1

    rot = math.radians(angle)
    sx = math.radians(shear[0])
    sy = math.radians(shear[1])

    cx, cy = center
    tx, ty = translate

    # RSS without scaling
    a = math.cos(rot - sy) / math.cos(sy)
    b = -math.cos(rot - sy) * math.tan(sx) / math.cos(sy) - math.sin(rot)
    c = math.sin(rot - sy) / math.cos(sy)
    d = -math.sin(rot - sy) * math.tan(sx) / math.cos(sy) + math.cos(rot)

    if inverted:
        # Inverted rotation matrix with scale and shear
        # det([[a, b], [c, d]]) == 1, since det(rotation) = 1 and det(shear) = 1
        matrix = [d, -b, 0.0, -c, a, 0.0]
        matrix = [x / scale for x in matrix]
        # Apply inverse of translation and of center translation: RSS^-1 * C^-1 * T^-1
        matrix[2] += matrix[0] * (-cx - tx) + matrix[1] * (-cy - ty)
        matrix[5] += matrix[3] * (-cx - tx) + matrix[4] * (-cy - ty)
        # Apply center translation: C * RSS^-1 * C^-1 * T^-1
        matrix[2] += cx
        matrix[5] += cy
    else:
        matrix = [a, b, 0.0, c, d, 0.0]
        matrix = [x * scale for x in matrix]
        # Apply inverse of center translation: RSS * C^-1
        matrix[2] += matrix[0] * (-cx) + matrix[1] * (-cy)
        matrix[5] += matrix[3] * (-cx) + matrix[4] * (-cy)
        # Apply translation and center : T * C * RSS * C^-1
        matrix[2] += cx + tx
        matrix[5] += cy + ty

    return matrix


def calc_centroid(mask: Tensor, id: int) -> Tensor:
    temp = (mask == id).float()

    # crop the region to just erode small region...
    lower = torch.nonzero(temp).min(0)[0]
    upper = torch.nonzero(temp).max(0)[0]

    temp = temp[
           lower[0].item(): upper[0].item(),  # x
           lower[1].item(): upper[1].item(),  # y
           lower[2].item(): upper[2].item(),  # z
           ]

    if temp.numel() == 0:
        return torch.tensor((-1, -1, -1), device=mask.device)

    x, y, z = temp.shape
    temp = temp.view((1, x, y, z))

    nonzero = torch.nonzero(temp)

    if nonzero.numel() == 0:
        return torch.tensor((-1, -1, -1), device=mask.device)

    old_temp = temp
    while nonzero.numel() > 0:
        old_temp = temp
        temp = temp.unsqueeze(0) if temp.ndim == 4 else temp
        temp = binary_erosion(temp)
        nonzero = torch.nonzero(temp)

    # except Exception
    #     print(temp.shape)
    #     print('threw error for some reason - merged_transform.py line 209')
    #     raise ValueError

    center = torch.nonzero(old_temp.view(x, y, z)).float().mean(0).add(lower)

    if torch.any(torch.isnan(center)):
        print(
            f"{id=}, {temp.shape=}, {old_temp.shape=}, {old_temp.numel()=}, {nonzero.numel()=}"
        )
        print(f"nonzero old_temp {torch.nonzero(old_temp.view(x, y, z)).float()}")
        print(upper, lower)
        raise ValueError

    return center


class TransformFromCfg(nn.Module):
    def __init__(self, cfg: CfgNode, device: torch.device, scale: float = 255.0):
        super(TransformFromCfg, self).__init__()
        """
        Why? Apparently a huge amount of overhead is just initializing this from cfg
        If we preinitalize, then we can save on overhead, to do this, we need a class...
        Probably a reasonalbe functional way to do this. Ill think on it later

        """

        self.prefix_function = self._identity
        self.posfix_function = self._identity

        self.dataset_mean = 0
        self.dataset_std = 1

        self.cfg = cfg

        self.DEVICE = device
        self.SCALE = scale

        self.CROP_WIDTH = cfg.AUGMENTATION.CROP_WIDTH
        self.CROP_HEIGHT = cfg.AUGMENTATION.CROP_HEIGHT

        self.CROP_DEPTH = cfg.AUGMENTATION.CROP_DEPTH

        self.FLIP_RATE = cfg.AUGMENTATION.FLIP_RATE

        self.BRIGHTNESS_RATE = cfg.AUGMENTATION.BRIGHTNESS_RATE
        self.BRIGHTNESS_RANGE = cfg.AUGMENTATION.BRIGHTNESS_RANGE
        self.NOISE_GAMMA = cfg.AUGMENTATION.NOISE_GAMMA
        self.NOISE_RATE = cfg.AUGMENTATION.NOISE_RATE

        self.FILTER_RATE = 0.5

        self.CONTRAST_RATE = cfg.AUGMENTATION.CONTRAST_RATE
        self.CONTRAST_RANGE = cfg.AUGMENTATION.CONTRAST_RANGE

        self.AFFINE_RATE = cfg.AUGMENTATION.AFFINE_RATE
        self.AFFINE_SCALE = cfg.AUGMENTATION.AFFINE_SCALE
        self.AFFINE_SHEAR = cfg.AUGMENTATION.AFFINE_SHEAR
        self.AFFINE_YAW = cfg.AUGMENTATION.AFFINE_YAW

        self.BAKE_SKELETON_ANISOTROPY = cfg.AUGMENTATION.BAKE_SKELETON_ANISOTROPY

        self._center = None
        self._xyz = (None, None, None)

    def _identity(self, *args):
        return args if len(args) > 1 else args[0]

    def _crop1(self, image, masks, skeletons):
        # ------------ Random Crop 1
        extra = 300
        C, X, Y, Z = image.shape
        # ------------ Random Crop 1
        extra = 300
        w = self.CROP_WIDTH + extra if self.CROP_WIDTH + extra <= X else X
        h = self.CROP_HEIGHT + extra if self.CROP_HEIGHT + extra <= Y else Y
        d = self.CROP_DEPTH if self.CROP_DEPTH <= Z else Z

        key = random.choice(list(skeletons.keys()))
        self._center: Tensor = skeletons[key].float().mean(0).squeeze()

        # Center that instance
        x0 = (
            self._center[0]
            .sub(w // 2)
            .long()
            .clamp(min=0, max=image.shape[1] - w)
        )
        y0 = (
            self._center[1]
            .sub(h // 2)
            .long()
            .clamp(min=0, max=image.shape[2] - h)
        )
        z0 = (
            self._center[2]
            .sub(d // 2)
            .long()
            .clamp(min=0, max=image.shape[3] - d)
        )

        self._xyz = (x0, y0, z0)

        x1 = x0 + w
        y1 = y0 + h
        z1 = z0 + d

        image = image[:, x0: x1, y0: y1, z0: z1]
        masks = masks[:, x0: x1, y0: y1, z0: z1]

        # Correct the skeleton positions
        unique = torch.unique(masks)
        if -1 not in skeletons:
            keys = list(skeletons.keys())
            for k in keys:
                if not torch.any(unique == k):
                    skeletons.pop(k)
                else:
                    skeletons[k] = skeletons[k] - torch.tensor([x0, y0, z0], device=self.DEVICE)

        # send to device
        if image.device != self.DEVICE:
            image = image.to(self.DEVICE)

        if masks.device != self.DEVICE:
            masks = masks.to(self.DEVICE)

        for k, v in skeletons.items():
            skeletons[k] = v.float().to(self.DEVICE)

        return image, masks, skeletons

    def _affine(self, image, masks, skeletons):
        angle = (self.AFFINE_YAW[1] - self.AFFINE_YAW[0]) * torch.rand(
            1, device=self.DEVICE
        ) + self.AFFINE_YAW[0]
        shear = (self.AFFINE_SHEAR[1] - self.AFFINE_SHEAR[0]) * torch.rand(
            1, device=self.DEVICE
        ) + self.AFFINE_SHEAR[0]
        scale = (self.AFFINE_SCALE[1] - self.AFFINE_SCALE[0]) * torch.rand(
            1, device=self.DEVICE
        ) + self.AFFINE_SCALE[0]

        # shear = torch.tensor((35), device=DEVICE)
        # angle = torch.tensor((45), device=DEVICE)
        # scale = torch.tensor((1), device=DEVICE)

        mat: Tensor = _get_affine_matrix(
            center=[image.shape[1] / 2, image.shape[2] / 2],
            angle=-angle.item(),
            translate=[0.0, 0.0],
            scale=scale.item(),
            shear=[0.0, float(shear.item())],
            device=str(image.device),
        )  # Rotate the skeletons by the affine matrix

        for k, v in skeletons.items():
            skeleton_xy = v[:, [0, 1]].permute(1, 0).unsqueeze(0)  # [N, 3] -> [1, 2, N]
            _ones = torch.ones(
                (1, 1, skeleton_xy.shape[-1]), device=self.DEVICE
            )  # [1, 1, N]
            skeleton_xy = torch.cat((skeleton_xy, _ones), dim=1)  # [1, 3, N]
            rotated_skeleton = mat @ skeleton_xy  # [1,3,N]
            skeletons[k][:, [0, 1]] = rotated_skeleton[0, [0, 1], :].T.float()

        image = ttf.affine(
            image.permute(0, 3, 1, 2).float(),
            angle=angle.item(),
            shear=float(shear.item()),
            scale=scale.item(),
            translate=[0, 0],
        ).permute(0, 2, 3, 1)

        # unique_before = masks.unique().long().sub(1)
        # unique_before = unique_before[unique_before.ge(0)]

        masks = ttf.affine(
            masks.permute(0, 3, 1, 2).float(),
            angle=angle.item(),
            shear=float(shear.item()),
            scale=scale.item(),
            translate=[0, 0],
        ).permute(0, 2, 3, 1)

        unique_after = masks.unique().long().sub(1)
        unique_after = unique_after[unique_after.ge(0)]

        skeletons = {
            k: v for k, v in skeletons.items() if torch.any(unique_after.eq(k - 1))
        }

        return image, masks, skeletons
        # angle = random.uniform(*self.AFFINE_YAW)
        # shear = random.uniform(*self.AFFINE_YAW)
        # scale = random.uniform(*self.AFFINE_SCALE)
        #
        # mat: Tensor = _get_affine_matrix(
        #     center=[image.shape[1] / 2, image.shape[2] / 2],
        #     angle=-angle,
        #     translate=[0.0, 0.0],
        #     scale=scale,
        #     shear=[0.0, shear],
        #     device=str(image.device),
        # )  # Rotate the skeletons by the affine matrix
        #
        # if -1 not in skeletons:
        #     for k, v in skeletons.items():
        #         skeleton_xy = v[:, [0, 1]].permute(1, 0).unsqueeze(0)  # [N, 3] -> [1, 2, N]
        #         _ones = torch.ones((1, 1, skeleton_xy.shape[-1]), device=self.DEVICE)  # [1, 1, N]
        #         skeleton_xy = torch.cat((skeleton_xy, _ones), dim=1)  # [1, 3, N]
        #         rotated_skeleton = mat @ skeleton_xy  # [1,3,N]
        #         skeletons[k][:, [0, 1]] = rotated_skeleton[0, [0, 1], :].T.float()
        #
        # image = ttf.affine(
        #     image.permute(0, 3, 1, 2).float(),
        #     angle=angle,
        #     shear=[float(shear)],
        #     scale=scale,
        #     translate=[0, 0],
        # ).permute(0, 2, 3, 1)
        #
        # masks = ttf.affine(
        #     masks.permute(0, 3, 1, 2).float(),
        #     angle=angle,
        #     shear=[float(shear)],
        #     scale=scale,
        #     translate=[0, 0],
        # ).permute(0, 2, 3, 1)
        #
        # return image, masks, skeletons

    def _crop2(self, image, masks, skeletons):
        C, X, Y, Z = image.shape
        w = self.CROP_WIDTH if self.CROP_WIDTH < X else X
        h = self.CROP_HEIGHT if self.CROP_HEIGHT < Y else Y
        d = self.CROP_DEPTH if self.CROP_DEPTH < Z else Z

        x0, y0, z0 = self._xyz

        self._center = self._center - torch.tensor([x0, y0, z0], device=self.DEVICE)

        # Center that instance
        x0 = (
            self._center[0]
            .sub(w // 2)
            .long()
            .clamp(min=0, max=image.shape[1] - w)
        )
        y0 = (
            self._center[1]
            .sub(h // 2)
            .long()
            .clamp(min=0, max=image.shape[2] - h)
        )
        z0 = (
            self._center[2]
            .sub(d // 2)
            .long()
            .clamp(min=0, max=image.shape[3] - d)
        )

        x1 = x0 + w
        y1 = y0 + h
        z1 = z0 + d

        image = image[:, x0:x1, y0:y1, z0:z1]
        masks = masks[:, x0:x1, y0:y1, z0:z1]

        unique = torch.unique(masks)
        if -1 not in skeletons:
            keys = list(skeletons.keys())
            for k in keys:
                if not torch.any(unique == k):
                    skeletons.pop(k)
                else:
                    skeletons[k] = skeletons[k] - torch.tensor([x0, y0, z0], device=self.DEVICE)

        return image, masks, skeletons

    def _flipX(self, image, masks, skeletons):
        image = image.flip(1)
        masks = masks.flip(1)

        if -1 not in skeletons:
            for k, v in skeletons.items():
                skeletons[k][:, 0] = image.shape[1] - v[:, 0]

        return image, masks

    def _flipY(self, image, masks, skeletons):
        image = image.flip(2)
        masks = masks.flip(2)
        if -1 not in skeletons:
            for k, v in skeletons.items():
                skeletons[k][:, 1] = image.shape[2] - v[:, 1]
        return image, masks

    def _flipZ(self, image, masks, skeletons):
        image = image.flip(3)
        masks = masks.flip(3)
        if -1 not in skeletons:
            for k, v in skeletons.items():
                skeletons[k][:, 2] = image.shape[3] - v[:, 2]
        return image, masks

    def _invert(self, image, masks):
        image.sub_(1).mul_(-1)
        return image, masks

    def _brightness(self, image, masks):
        val = random.uniform(*self.BRIGHTNESS_RANGE)
        # in place ok because flip always returns a copy
        image = image.add(val)
        return image, masks

    def _contrast(self, image, masks):
        contrast_val = random.uniform(*self.CONTRAST_RANGE)
        # [ C, X, Y, Z ] -> [Z, C, X, Y]
        image = ttf.adjust_contrast(image.permute(3, 0, 1, 2), contrast_val).permute(1, 2, 3, 0)

        return image, masks

    def _noise(self, image, masks):
        noise = torch.rand(image.shape, device=self.DEVICE) * self.NOISE_GAMMA
        image = image.add(noise)
        return image, masks

    def _normalize(self, image, masks):
        # mean = image.float().mean()
        # std = image.float().std()
        mean = image.float().mean() if not self.dataset_mean else self.dataset_mean
        std = image.float().std() if not self.dataset_std else self.dataset_std

        image = image.float().sub(mean).div(std)
        return image, masks

    def set_dataset_mean(self, mean):
        self.dataset_mean = mean
        return self

    def set_dataset_std(self, std):
        self.dataset_std = std
        return self

    @torch.no_grad()
    def forward(self, data_dict: Dict[str, Tensor]) -> Dict[str, Tensor]:

        assert "masks" in data_dict, 'keyword "masks" not in data_dict'
        assert "image" in data_dict, 'keyword "image" not in data_dict'
        assert "skeletons" in data_dict, 'keyword "skeletons" not in data_dict'

        logging.debug(f"TransformFromCfg.forward() | starting transforms on device: {self.DEVICE}")


        logging.debug("TransformFromCfg.forward() | applying prefix function")
        data_dict = self.prefix_function(data_dict)

        masks = data_dict["masks"]
        image = data_dict["image"]
        skeletons = deepcopy(data_dict["skeletons"])

        spatial_dims = masks.ndim - 1

        logging.debug("TransformFromCfg.forward() | applying crop 1")
        image, masks, skeletons = self._crop1(image, masks, skeletons)

        # scale: int = 2 ** 16 if image.max() > 256 else 255  # Our images might be 16 bit, or 8 bit
        # scale = scale if image.max() > 1 else 1.0

        logging.debug("TransformFromCfg.forward() | normalizing")
        image, masks = self._normalize(image, masks)

        # affine
        if random.random() < self.AFFINE_RATE:
            logging.debug("TransformFromCfg.forward() | applying affine transform")
            image, masks, skeletons = self._affine(image, masks, skeletons)

        # ------------ Center Crop 2
        logging.debug("TransformFromCfg.forward() | applying crop 2")
        image, masks, skeletons = self._crop2(image, masks, skeletons)

        # ------------------- x flip
        if random.random() < self.FLIP_RATE:
            logging.debug("TransformFromCfg.forward() | flipping in x")
            image, masks = self._flipX(image, masks, skeletons)

        # ------------------- y flip
        if random.random() < self.FLIP_RATE:
            logging.debug("TransformFromCfg.forward() | flipping in y")
            image, masks = self._flipY(image, masks, skeletons)

        # ------------------- z flip
        if random.random() < self.FLIP_RATE:
            logging.debug("TransformFromCfg.forward() | flipping in z")
            image, masks = self._flipZ(image, masks, skeletons)

        # # ------------------- Random Invert
        if random.random() < self.BRIGHTNESS_RATE:
            logging.debug("TransformFromCfg.forward() | inverting")
            image, masks = self._invert(image, masks)

        # ------------------- Adjust Brightness
        if random.random() < self.BRIGHTNESS_RATE:
            logging.debug("TransformFromCfg.forward() | adjusting brightness")
            image, masks = self._brightness(image, masks)

        # ------------------- Adjust Contrast
        if random.random() < self.CONTRAST_RATE:
            logging.debug("TransformFromCfg.forward() | adjusting contrast")
            image, masks = self._contrast(image, masks)

        # ------------------- Noise
        if random.random() < self.NOISE_RATE:
            logging.debug("TransformFromCfg.forward() | adding noise")
            image, masks = self._noise(image, masks)

        if spatial_dims == 2:
            image = image[..., 0]
            masks = masks[..., 0]

        data_dict["image"] = image
        data_dict["masks"] = masks

        baked: Tensor = bake_skeleton(
            masks,
            skeletons,
            anisotropy=self.BAKE_SKELETON_ANISOTROPY,
            average=True,
            device=self.DEVICE,
        )
        data_dict["baked_skeleton"]: Union[Tensor, None] = baked

        _, x, y, z = masks.shape
        data_dict["skele_masks"]: Tensor = skeleton_to_mask(
            skeletons,
            (x, y, z),
            device=self.DEVICE
            # kernel_size=cfg.AUGMENTATION.SMOOTH_SKELETON_KERNEL_SIZE,
            # n=cfg.AUGMENTATION.N_SKELETON_MASK_DILATE,
        )

        data_dict = self.posfix_function(data_dict)

        return data_dict

    def pre_fn(self, fn: Callable[[Dict[str, Tensor]], Dict[str, Tensor]]):
        self.prefix_function = fn
        return self

    def post_fn(self, fn: Callable[[Dict[str, Tensor]], Dict[str, Tensor]]):
        self.posfix_function = fn
        return self

    def post_crop_fn(self, fn):
        self.postcrop_function = fn
        return self

    def __repr__(self):
        return f"TransformFromCfg[Device:{self.DEVICE}]\ncfg.AUGMENTATION:\n=================\n{self.cfg.AUGMENTATION}]"

#
# @torch.no_grad()
# # @torch.jit.ignore()
# def transform_from_cfg(
#         data_dict: Dict[str, Tensor], cfg: CfgNode, device: Optional[str] = None
# ) -> DataDict:
#     DEVICE: str = str(data_dict["image"].device) if device is None else device
#
#     # Image should be in shape of [C, H, W, D]
#     CROP_WIDTH = torch.tensor(cfg.AUGMENTATION.CROP_WIDTH, device=DEVICE)
#     CROP_HEIGHT = torch.tensor(cfg.AUGMENTATION.CROP_HEIGHT, device=DEVICE)
#     CROP_DEPTH = torch.tensor(cfg.AUGMENTATION.CROP_DEPTH, device=DEVICE)
#
#     FLIP_RATE = torch.tensor(cfg.AUGMENTATION.FLIP_RATE, device=DEVICE)
#
#     BRIGHTNESS_RATE = torch.tensor(cfg.AUGMENTATION.BRIGHTNESS_RATE, device=DEVICE)
#     BRIGHTNESS_RANGE = torch.tensor(cfg.AUGMENTATION.BRIGHTNESS_RANGE, device=DEVICE)
#
#     NOISE_GAMMA = torch.tensor(cfg.AUGMENTATION.NOISE_GAMMA, device=DEVICE)
#     NOISE_RATE = torch.tensor(cfg.AUGMENTATION.NOISE_RATE, device=DEVICE)
#
#     FILTER_RATE = torch.tensor(0.5, device=DEVICE)
#
#     CONTRAST_RATE = torch.tensor(cfg.AUGMENTATION.CONTRAST_RATE, device=DEVICE)
#     CONTRAST_RANGE = torch.tensor(cfg.AUGMENTATION.CONTRAST_RANGE, device=DEVICE)
#
#     AFFINE_RATE = torch.tensor(cfg.AUGMENTATION.AFFINE_RATE, device=DEVICE)
#     AFFINE_SCALE = torch.tensor(cfg.AUGMENTATION.AFFINE_SCALE, device=DEVICE)
#     AFFINE_YAW = torch.tensor(cfg.AUGMENTATION.AFFINE_YAW, device=DEVICE)
#     AFFINE_SHEAR = torch.tensor(cfg.AUGMENTATION.AFFINE_SHEAR, device=DEVICE)
#
#     masks = data_dict["masks"]  # .to(DEVI
#     image = data_dict["image"]
#     #
#     skeletons = deepcopy(data_dict["skeletons"])
#     skeletons = {k: v.float() for k, v in skeletons.items()}
#
#     # ------------ Random Crop 1
#     extra = 300
#     w = (
#         CROP_WIDTH + extra
#         if CROP_WIDTH + extra <= image.shape[1]
#         else torch.tensor(image.shape[1])
#     )
#     h = (
#         CROP_HEIGHT + extra
#         if CROP_HEIGHT + extra <= image.shape[2]
#         else torch.tensor(image.shape[2])
#     )
#     d = CROP_DEPTH if CROP_DEPTH <= image.shape[3] else torch.tensor(image.shape[3])
#
#     # Randomly select a centroid to center in frame
#     ind: int = torch.randint(
#         len(skeletons.keys()), (1,), dtype=torch.long, device=DEVICE
#     ).item()
#     key: int = list(skeletons)[ind]
#     center: Tensor = skeletons[key].mean(0).squeeze()
#
#     # Center that instance
#     x0 = (
#         center[0]
#         .sub(torch.floor(w / 2))
#         .long()
#         .clamp(min=0, max=image.shape[1] - w.item())
#     )
#     y0 = (
#         center[1]
#         .sub(torch.floor(h / 2))
#         .long()
#         .clamp(min=0, max=image.shape[2] - h.item())
#     )
#     z0 = (
#         center[2]
#         .sub(torch.floor(d / 2))
#         .long()
#         .clamp(min=0, max=image.shape[3] - d.item())
#     )
#
#     x1 = x0 + w
#     y1 = y0 + h
#     z1 = z0 + d
#
#     image = (
#         image[:, x0.item(): x1.item(), y0.item(): y1.item(), z0.item(): z1.item()]
#         .to(DEVICE)
#         .div(255)
#         .half()
#     )
#
#     masks = masks[
#             :, x0.item(): x1.item(), y0.item(): y1.item(), z0.item(): z1.item()
#             ].to(DEVICE)
#
#     # Correct the skeleton positions
#     unique = torch.unique(masks)
#     if -1 not in skeletons:
#         keys = list(skeletons.keys())
#         for k in keys:
#             if not torch.any(unique == k):
#                 skeletons.pop(k)
#             else:
#                 skeletons[k] = skeletons[k].to(DEVICE) - torch.tensor([x0, y0, z0], device=DEVICE)
#
#     # --------------------------- elastic transform
#     # image, masks, skeletons = elastic_deform(image.unsqueeze(0), masks.unsqueeze(0), skeletons)
#     # image = image.squeeze(0)
#     # masks = masks.squeeze(0)
#
#     # -------------------affine (Cant use baked skeletons)
#     if torch.rand(1, device=DEVICE) < AFFINE_RATE:
#         angle = (AFFINE_YAW[1] - AFFINE_YAW[0]) * torch.rand(
#             1, device=DEVICE
#         ) + AFFINE_YAW[0]
#         shear = (AFFINE_SHEAR[1] - AFFINE_SHEAR[0]) * torch.rand(
#             1, device=DEVICE
#         ) + AFFINE_SHEAR[0]
#         scale = (AFFINE_SCALE[1] - AFFINE_SCALE[0]) * torch.rand(
#             1, device=DEVICE
#         ) + AFFINE_SCALE[0]
#
#         # shear = torch.tensor((35), device=DEVICE)
#         # angle = torch.tensor((45), device=DEVICE)
#         # scale = torch.tensor((1), device=DEVICE)
#
#         mat: Tensor = _get_affine_matrix(
#             center=[image.shape[1] / 2, image.shape[2] / 2],
#             angle=-angle.item(),
#             translate=[0.0, 0.0],
#             scale=scale.item(),
#             shear=[0.0, float(shear.item())],
#             device=str(image.device),
#         )  # Rotate the skeletons by the affine matrix
#
#         for k, v in skeletons.items():
#             skeleton_xy = v[:, [0, 1]].permute(1, 0).unsqueeze(0)  # [N, 3] -> [1, 2, N]
#             _ones = torch.ones(
#                 (1, 1, skeleton_xy.shape[-1]), device=DEVICE
#             )  # [1, 1, N]
#             skeleton_xy = torch.cat((skeleton_xy, _ones), dim=1)  # [1, 3, N]
#             rotated_skeleton = mat @ skeleton_xy  # [1,3,N]
#             skeletons[k][:, [0, 1]] = rotated_skeleton[0, [0, 1], :].T.float()
#
#         image = ttf.affine(
#             image.permute(0, 3, 1, 2).float(),
#             angle=angle.item(),
#             shear=float(shear.item()),
#             scale=scale.item(),
#             translate=[0, 0],
#         ).permute(0, 2, 3, 1)
#
#         # unique_before = masks.unique().long().sub(1)
#         # unique_before = unique_before[unique_before.ge(0)]
#
#         masks = ttf.affine(
#             masks.permute(0, 3, 1, 2).float(),
#             angle=angle.item(),
#             shear=float(shear.item()),
#             scale=scale.item(),
#             translate=[0, 0],
#         ).permute(0, 2, 3, 1)
#
#         unique_after = masks.unique().long().sub(1)
#         unique_after = unique_after[unique_after.ge(0)]
#
#         skeletons = {
#             k: v for k, v in skeletons.items() if torch.any(unique_after.eq(k - 1))
#         }
#
#         # assert len(skeletons.keys()) > 0, f"{unique_after=}, {unique_before=}"
#
#     # # ------------ Center Crop 2
#     w = CROP_WIDTH if CROP_WIDTH <= image.shape[1] else torch.tensor(image.shape[1])
#     h = CROP_HEIGHT if CROP_HEIGHT <= image.shape[2] else torch.tensor(image.shape[2])
#     d = CROP_DEPTH if CROP_DEPTH <= image.shape[3] else torch.tensor(image.shape[3])
#
#     center = center - torch.tensor([x0, y0, z0], device=DEVICE)
#
#     # Center that instance
#     x0 = (
#         center[0]
#         .sub(torch.floor(w / 2))
#         .long()
#         .clamp(min=0, max=image.shape[1] - w.item())
#     )
#     y0 = (
#         center[1]
#         .sub(torch.floor(h / 2))
#         .long()
#         .clamp(min=0, max=image.shape[2] - h.item())
#     )
#     z0 = (
#         center[2]
#         .sub(torch.floor(d / 2))
#         .long()
#         .clamp(min=0, max=image.shape[3] - d.item())
#     )
#
#     x1 = x0 + w
#     y1 = y0 + h
#     z1 = z0 + d
#
#     image = image[
#             :, x0.item(): x1.item(), y0.item(): y1.item(), z0.item(): z1.item()
#             ]
#     masks = masks[
#             :, x0.item(): x1.item(), y0.item(): y1.item(), z0.item(): z1.item()
#             ]
#
#     unique = torch.unique(masks)
#     if -1 not in skeletons:
#         keys = list(skeletons.keys())
#         for k in keys:
#             if not torch.any(unique == k):
#                 skeletons.pop(k)
#             else:
#                 skeletons[k] = skeletons[k] - torch.tensor([x0, y0, z0], device=DEVICE)
#
#     # ------------------- x flip
#     if torch.rand(1, device=DEVICE) < FLIP_RATE:
#         image = image.flip(1)
#         masks = masks.flip(1)
#
#         if -1 not in skeletons:
#             for k, v in skeletons.items():
#                 skeletons[k][:, 0] = image.shape[1] - v[:, 0]
#
#     # ------------------- y flip
#     if torch.rand(1, device=DEVICE) < FLIP_RATE:
#         image = image.flip(2)
#         masks = masks.flip(2)
#         if -1 not in skeletons:
#             for k, v in skeletons.items():
#                 skeletons[k][:, 1] = image.shape[2] - v[:, 1]
#
#     # ------------------- z flip
#     if torch.rand(1, device=DEVICE) < FLIP_RATE:
#         image = image.flip(3)
#         masks = masks.flip(3)
#         if -1 not in skeletons:
#             for k, v in skeletons.items():
#                 skeletons[k][:, 2] = image.shape[3] - v[:, 2]
#
#     # # ------------------- Random Invert
#     if torch.rand(1, device=DEVICE) < BRIGHTNESS_RATE:
#         image = image.sub(1).mul(-1)
#
#     # ------------------- Adjust Brightness
#     if torch.rand(1, device=DEVICE) < BRIGHTNESS_RATE:
#         # funky looking but FAST
#         val = torch.empty(image.shape[0], device=DEVICE).uniform_(
#             BRIGHTNESS_RANGE[0], BRIGHTNESS_RANGE[1]
#         )
#         image = image.add(val.reshape(image.shape[0], 1, 1, 1)).clamp(0, 1)
#
#     # ------------------- Adjust Contrast
#     if torch.rand(1, device=DEVICE) < CONTRAST_RATE:
#         contrast_val = (CONTRAST_RANGE[1] - CONTRAST_RANGE[0]) * torch.rand(
#             (image.shape[0]), device=DEVICE
#         ) + CONTRAST_RANGE[0]
#
#         for z in range(image.shape[-1]):
#             image[..., z] = ttf.adjust_contrast(
#                 image[..., z], contrast_val[0].item()
#             ).squeeze(0)
#
#     # ------------------- Noise
#     if torch.rand(1, device=DEVICE) < NOISE_RATE:
#         noise = torch.rand(image.shape, device=DEVICE) * NOISE_GAMMA
#
#         image = image.add(noise).clamp(0, 1)
#     #
#     data_dict["image"] = image
#     data_dict["masks"] = masks
#     data_dict["skeletons"] = skeletons
#
#     baked: Tensor = bake_skeleton(
#         masks,
#         skeletons,
#         anisotropy=cfg.AUGMENTATION.BAKE_SKELETON_ANISOTROPY,
#         average=True,
#         device=DEVICE,
#     )
#     data_dict["baked_skeleton"]: Union[Tensor, None] = baked
#
#     _, x, y, z = masks.shape
#     data_dict["skele_masks"]: Tensor = skeleton_to_mask(
#         skeletons,
#         (x, y, z),
#         # kernel_size=cfg.AUGMENTATION.SMOOTH_SKELETON_KERNEL_SIZE,
#         # n=cfg.AUGMENTATION.N_SKELETON_MASK_DILATE,
#     )
#
#     return data_dict

class BackgroundTransformFromCfg(TransformFromCfg):
    def __init__(self, cfg: CfgNode, device: torch.device | str):
        super(BackgroundTransformFromCfg, self).__init__(cfg, device)

    def _crop1(self, image, masks, skeletons):
        masks = masks.unsqueeze(-1) if masks.ndim == 3 else masks
        image = image.unsqueeze(-1) if image.ndim == 3 else image

        C, X, Y, Z = image.shape
        # ------------ Random Crop 1
        extra = 300
        w = self.CROP_WIDTH + extra if self.CROP_WIDTH + extra <= X else X
        h = self.CROP_HEIGHT + extra if self.CROP_HEIGHT + extra <= Y else Y
        d = self.CROP_DEPTH if self.CROP_DEPTH <= Z else Z

        # select a random point for croping
        x0 = random.randint(0, X - w)
        y0 = random.randint(0, Y - h)
        z0 = random.randint(0, Z - d)

        x1 = x0 + w
        y1 = y0 + h
        z1 = z0 + d

        image = image[:, x0:x1, y0:y1, z0:z1]
        masks = masks[:, x0:x1, y0:y1, z0:z1]

        if image.device != self.DEVICE:
            image = image.to(self.DEVICE)

        if masks.device != self.DEVICE:
            masks = masks.to(self.DEVICE)

        return image, masks, skeletons

    def forward(self, data_dict):
        data_dict['masks'] = torch.ones_like(data_dict['image'])
        data_dict['skeletons'] = {-1: None}

        data_dict = super().forward(data_dict)
        return

    def __repr__(self):
        return f"BackgroundTransformFromCfg[Device:{self.DEVICE}]\ncfg.AUGMENTATION:\n=================\n{self.cfg.AUGMENTATION}]"

#
# def background_transform_from_cfg(
#         data_dict: Dict[str, Tensor], cfg: CfgNode, device: Optional[str] = None
# ) -> DataDict:
#     # Image should be in shape of [C, H, W, D]
#     DEVICE: str = str(data_dict["image"].device) if device is None else device
#
#     # Image should be in shape of [C, H, W, D]
#     CROP_WIDTH = torch.tensor(cfg.AUGMENTATION.CROP_WIDTH, device=DEVICE)
#     CROP_HEIGHT = torch.tensor(cfg.AUGMENTATION.CROP_HEIGHT, device=DEVICE)
#     CROP_DEPTH = torch.tensor(cfg.AUGMENTATION.CROP_DEPTH, device=DEVICE)
#
#     FLIP_RATE = torch.tensor(cfg.AUGMENTATION.FLIP_RATE, device=DEVICE)
#
#     BRIGHTNESS_RATE = torch.tensor(cfg.AUGMENTATION.BRIGHTNESS_RATE, device=DEVICE)
#     BRIGHTNESS_RANGE = torch.tensor(cfg.AUGMENTATION.BRIGHTNESS_RANGE, device=DEVICE)
#
#     NOISE_GAMMA = torch.tensor(cfg.AUGMENTATION.NOISE_GAMMA, device=DEVICE)
#     NOISE_RATE = torch.tensor(cfg.AUGMENTATION.NOISE_RATE, device=DEVICE)
#
#     FILTER_RATE = torch.tensor(0.5, device=DEVICE)
#
#     CONTRAST_RATE = torch.tensor(cfg.AUGMENTATION.CONTRAST_RATE, device=DEVICE)
#     CONTRAST_RANGE = torch.tensor(cfg.AUGMENTATION.CONTRAST_RANGE, device=DEVICE)
#
#     AFFINE_RATE = torch.tensor(cfg.AUGMENTATION.AFFINE_RATE, device=DEVICE)
#     AFFINE_SCALE = torch.tensor(cfg.AUGMENTATION.AFFINE_SCALE, device=DEVICE)
#     AFFINE_YAW = torch.tensor(cfg.AUGMENTATION.AFFINE_YAW, device=DEVICE)
#     AFFINE_SHEAR = torch.tensor(cfg.AUGMENTATION.AFFINE_SHEAR, device=DEVICE)
#
#     image = torch.clone(data_dict["image"])  #
#
#     # ------------ Random Crop 1
#     extra = 300
#     w = (
#         CROP_WIDTH + extra
#         if CROP_WIDTH + extra <= image.shape[1]
#         else torch.tensor(image.shape[1])
#     )
#     h = (
#         CROP_HEIGHT + extra
#         if CROP_HEIGHT + extra <= image.shape[2]
#         else torch.tensor(image.shape[2])
#     )
#     d = CROP_DEPTH if CROP_DEPTH <= image.shape[3] else torch.tensor(image.shape[3])
#
#     shape = torch.tensor(image.shape[1::], device=DEVICE) - torch.tensor(
#         [w, h, d], device=DEVICE
#     )
#     center = torch.tensor(
#         [torch.randint(0, s, (1,)).item() if s > 0 else 0 for s in shape], device=DEVICE
#     )
#
#     # Center that instance
#     x0 = (
#         center[0]
#         .sub(torch.floor(w / 2))
#         .long()
#         .clamp(min=0, max=image.shape[1] - w.item())
#     )
#     y0 = (
#         center[1]
#         .sub(torch.floor(h / 2))
#         .long()
#         .clamp(min=0, max=image.shape[2] - h.item())
#     )
#     z0 = (
#         center[2]
#         .sub(torch.floor(d / 2))
#         .long()
#         .clamp(min=0, max=image.shape[3] - d.item())
#     )
#
#     x1 = x0 + w
#     y1 = y0 + h
#     z1 = z0 + d
#
#     image = image[
#             :, x0.item(): x1.item(), y0.item(): y1.item(), z0.item(): z1.item()
#             ].to(DEVICE)
#
#     # -------------------affine (Cant use baked skeletons)
#     if torch.rand(1, device=DEVICE) < AFFINE_RATE:
#         angle = (AFFINE_YAW[1] - AFFINE_YAW[0]) * torch.rand(
#             1, device=DEVICE
#         ) + AFFINE_YAW[0]
#         shear = (AFFINE_SHEAR[1] - AFFINE_SHEAR[0]) * torch.rand(
#             1, device=DEVICE
#         ) + AFFINE_SHEAR[0]
#         scale = (AFFINE_SCALE[1] - AFFINE_SCALE[0]) * torch.rand(
#             1, device=DEVICE
#         ) + AFFINE_SCALE[0]
#
#         image = ttf.affine(
#             image.permute(0, 3, 1, 2).float(),
#             angle=angle.item(),
#             shear=[float(shear.item())],
#             scale=scale.item(),
#             translate=[0, 0],
#         ).permute(0, 2, 3, 1)
#
#     # ------------ Center Crop 2
#     w = CROP_WIDTH if CROP_WIDTH <= image.shape[1] else torch.tensor(image.shape[1])
#     h = CROP_HEIGHT if CROP_HEIGHT <= image.shape[2] else torch.tensor(image.shape[2])
#     d = CROP_DEPTH if CROP_DEPTH <= image.shape[3] else torch.tensor(image.shape[3])
#
#     center = center - torch.tensor([x0, y0, z0], device=DEVICE)
#
#     # Center that instance
#     x0 = (
#         center[0]
#         .sub(torch.floor(w / 2))
#         .long()
#         .clamp(min=0, max=image.shape[1] - w.item())
#     )
#     y0 = (
#         center[1]
#         .sub(torch.floor(h / 2))
#         .long()
#         .clamp(min=0, max=image.shape[2] - h.item())
#     )
#     z0 = (
#         center[2]
#         .sub(torch.floor(d / 2))
#         .long()
#         .clamp(min=0, max=image.shape[3] - d.item())
#     )
#
#     x1 = x0 + w
#     y1 = y0 + h
#     z1 = z0 + d
#
#     image = image[
#             :, x0.item(): x1.item(), y0.item(): y1.item(), z0.item(): z1.item()
#             ]
#
#     # ------------------- x flip
#     if torch.rand(1, device=DEVICE) < FLIP_RATE:
#         image = image.flip(1)
#
#     # ------------------- y flip
#     if torch.rand(1, device=DEVICE) < FLIP_RATE:
#         image = image.flip(2)
#
#     # ------------------- z flip
#     if torch.rand(1, device=DEVICE) < FLIP_RATE:
#         image = image.flip(3)
#
#     # # ------------------- Random Invert
#     if torch.rand(1, device=DEVICE) < BRIGHTNESS_RATE:
#         image = image.sub(1).mul(-1)
#
#     # ------------------- Adjust Brightness
#     if torch.rand(1, device=DEVICE) < BRIGHTNESS_RATE:
#         # funky looking but FAST
#         val = torch.empty(image.shape[0], device=DEVICE).uniform_(
#             BRIGHTNESS_RANGE[0], BRIGHTNESS_RANGE[1]
#         )
#         image = image.add(val.reshape(image.shape[0], 1, 1, 1)).clamp(0, 1)
#
#     # ------------------- Adjust Contrast
#     if torch.rand(1, device=DEVICE) < CONTRAST_RATE:
#         contrast_val = (CONTRAST_RANGE[1] - CONTRAST_RANGE[0]) * torch.rand(
#             (image.shape[0]), device=DEVICE
#         ) + CONTRAST_RANGE[0]
#
#         for z in range(image.shape[-1]):
#             image[..., z] = ttf.adjust_contrast(
#                 image[..., z], contrast_val[0].item()
#             ).squeeze(0)
#
#     # ------------------- Noise
#     if torch.rand(1, device=DEVICE) < NOISE_RATE:
#         noise = torch.rand(image.shape, device=DEVICE) * NOISE_GAMMA
#
#         image = image.add(noise).clamp(0, 1)
#
#     data_dict["image"] = image
#     data_dict["masks"] = torch.zeros_like(image, device=DEVICE)
#     data_dict["skeletons"]: Dict[int, Tensor] = {-1: torch.empty((0, 3), device=DEVICE)}
#     data_dict["baked_skeleton"] = torch.zeros(
#         (3, image.shape[1], image.shape[2], image.shape[3]), device=DEVICE
#     )
#     data_dict["skele_masks"] = torch.zeros_like(image, device=DEVICE)
#
#
#     return data_dict

#
# @torch.no_grad()
# def merged_transform_3D(
#         data_dict: Dict[str, Tensor],
#         device: Optional[str] = None,
#         bake_skeleton_anisotropy: Tuple[float, float, float] = (1.0, 1.0, 3.0),
#         smooth_skeleton_kernel_size: Tuple[int, int, int] = (3, 3, 1),
# ) -> DataDict:
#     DEVICE: str = str(data_dict["image"].device) if device is None else device
#
#     # Image should be in shape of [C, H, W, D]
#     CROP_WIDTH = torch.tensor([300], device=DEVICE)
#     CROP_HEIGHT = torch.tensor([300], device=DEVICE)
#     CROP_DEPTH = torch.tensor([20], device=DEVICE)
#
#     FLIP_RATE = torch.tensor(0.5, device=DEVICE)
#
#     BRIGHTNESS_RATE = torch.tensor(0.4, device=DEVICE)
#     BRIGHTNESS_RANGE = torch.tensor((-0.1, 0.1), device=DEVICE)
#
#     NOISE_GAMMA = torch.tensor(0.1, device=DEVICE)
#     NOISE_RATE = torch.tensor(0.2, device=DEVICE)
#
#     FILTER_RATE = torch.tensor(0.5, device=DEVICE)
#
#     CONTRAST_RATE = torch.tensor(0.33, device=DEVICE)
#     CONTRAST_RANGE = torch.tensor((0.75, 2.0), device=DEVICE)
#
#     AFFINE_RATE = torch.tensor(0.66, device=DEVICE)
#     AFFINE_SCALE = torch.tensor((0.85, 1.1), device=DEVICE)
#     AFFINE_YAW = torch.tensor((-180, 180), device=DEVICE)
#     AFFINE_SHEAR = torch.tensor((-7, 7), device=DEVICE)
#
#     masks = torch.clone(data_dict["masks"])  # .to(DEVICE))
#     image = torch.clone(data_dict["image"])  #
#
#     skeletons = deepcopy(data_dict["skeletons"])
#     skeletons = {k: v.float().to(DEVICE) for k, v in skeletons.items()}
#
#     # ------------ Random Crop 1
#     extra = 300
#     w = (
#         CROP_WIDTH + extra
#         if CROP_WIDTH + extra <= image.shape[1]
#         else torch.tensor(image.shape[1])
#     )
#     h = (
#         CROP_HEIGHT + extra
#         if CROP_HEIGHT + extra <= image.shape[2]
#         else torch.tensor(image.shape[2])
#     )
#     d = CROP_DEPTH if CROP_DEPTH <= image.shape[3] else torch.tensor(image.shape[3])
#
#     # Randomly select a centroid to center in frame
#     ind: int = torch.randint(
#         len(skeletons.keys()), (1,), dtype=torch.long, device=DEVICE
#     ).item()
#     key: int = list(skeletons)[ind]
#     center: Tensor = skeletons[key].mean(0).squeeze()
#
#     # Center that instance
#     x0 = (
#         center[0]
#         .sub(torch.floor(w / 2))
#         .long()
#         .clamp(min=0, max=image.shape[1] - w.item())
#     )
#     y0 = (
#         center[1]
#         .sub(torch.floor(h / 2))
#         .long()
#         .clamp(min=0, max=image.shape[2] - h.item())
#     )
#     z0 = (
#         center[2]
#         .sub(torch.floor(d / 2))
#         .long()
#         .clamp(min=0, max=image.shape[3] - d.item())
#     )
#
#     x1 = x0 + w
#     y1 = y0 + h
#     z1 = z0 + d
#
#     image = image[
#             :, x0.item(): x1.item(), y0.item(): y1.item(), z0.item(): z1.item()
#             ].to(DEVICE)
#     masks = masks[
#             :, x0.item(): x1.item(), y0.item(): y1.item(), z0.item(): z1.item()
#             ].to(DEVICE)
#
#     # Correct the skeleton positions
#     unique = torch.unique(masks)
#     if -1 not in skeletons:
#         keys = list(skeletons.keys())
#         for k in keys:
#             if not torch.any(unique == k):
#                 skeletons.pop(k)
#             else:
#                 skeletons[k] = skeletons[k] - torch.tensor([x0, y0, z0], device=DEVICE)
#
#     # -------------------affine (Cant use baked skeletons)
#     if torch.rand(1, device=DEVICE) < AFFINE_RATE:
#         angle = (AFFINE_YAW[1] - AFFINE_YAW[0]) * torch.rand(
#             1, device=DEVICE
#         ) + AFFINE_YAW[0]
#         shear = (AFFINE_SHEAR[1] - AFFINE_SHEAR[0]) * torch.rand(
#             1, device=DEVICE
#         ) + AFFINE_SHEAR[0]
#         scale = (AFFINE_SCALE[1] - AFFINE_SCALE[0]) * torch.rand(
#             1, device=DEVICE
#         ) + AFFINE_SCALE[0]
#
#         mat: Tensor = _get_affine_matrix(
#             center=[image.shape[1] / 2, image.shape[2] / 2],
#             angle=-angle.item(),
#             translate=[0.0, 0.0],
#             scale=scale.item(),
#             shear=[float(shear.item()), float(shear.item())],
#             device=str(image.device),
#         )
#
#         # Rotate the skeletons by the affine matrix
#         for k, v in skeletons.items():
#             skeleton_xy = v[:, [0, 1]].permute(1, 0).unsqueeze(0)  # [N, 3] -> [1, 2, N]
#             _ones = torch.ones(
#                 (1, 1, skeleton_xy.shape[-1]), device=DEVICE
#             )  # [1, 1, N]
#             skeleton_xy = torch.cat((skeleton_xy, _ones), dim=1)  # [1, 3, N]
#             rotated_skeleton = mat @ skeleton_xy  # [1,3,N]
#             skeletons[k][:, [0, 1]] = rotated_skeleton[0, [0, 1], :].T.float()
#
#         image = ttf.affine(
#             image.permute(0, 3, 1, 2).float(),
#             angle=angle.item(),
#             shear=[float(shear.item())],
#             scale=scale.item(),
#             translate=[0, 0],
#         ).permute(0, 2, 3, 1)
#
#         unique_before = masks.unique().long().sub(1)
#         unique_before = unique_before[unique_before.ge(0)]
#
#         masks = ttf.affine(
#             masks.permute(0, 3, 1, 2).float(),
#             angle=angle.item(),
#             shear=[float(shear.item())],
#             scale=scale.item(),
#             translate=[0, 0],
#         ).permute(0, 2, 3, 1)
#
#         unique_after = masks.unique().long().sub(1)
#         unique_after = unique_after[unique_after.ge(0)]
#
#         skeletons = {
#             k: v for k, v in skeletons.items() if torch.any(unique_after.eq(k - 1))
#         }
#
#         assert len(skeletons.keys()) > 0, f"{unique_after=}, {unique_before=}"
#
#     # ------------ Center Crop 2
#     w = CROP_WIDTH if CROP_WIDTH <= image.shape[1] else torch.tensor(image.shape[1])
#     h = CROP_HEIGHT if CROP_HEIGHT <= image.shape[2] else torch.tensor(image.shape[2])
#     d = CROP_DEPTH if CROP_DEPTH <= image.shape[3] else torch.tensor(image.shape[3])
#
#     center = center - torch.tensor([x0, y0, z0], device=DEVICE)
#
#     # Center that instance
#     x0 = (
#         center[0]
#         .sub(torch.floor(w / 2))
#         .long()
#         .clamp(min=0, max=image.shape[1] - w.item())
#     )
#     y0 = (
#         center[1]
#         .sub(torch.floor(h / 2))
#         .long()
#         .clamp(min=0, max=image.shape[2] - h.item())
#     )
#     z0 = (
#         center[2]
#         .sub(torch.floor(d / 2))
#         .long()
#         .clamp(min=0, max=image.shape[3] - d.item())
#     )
#
#     x1 = x0 + w
#     y1 = y0 + h
#     z1 = z0 + d
#
#     image = image[
#             :, x0.item(): x1.item(), y0.item(): y1.item(), z0.item(): z1.item()
#             ]
#     masks = masks[
#             :, x0.item(): x1.item(), y0.item(): y1.item(), z0.item(): z1.item()
#             ]
#
#     unique = torch.unique(masks)
#     if -1 not in skeletons:
#         keys = list(skeletons.keys())
#         for k in keys:
#             if not torch.any(unique == k):
#                 skeletons.pop(k)
#             else:
#                 skeletons[k] = skeletons[k] - torch.tensor([x0, y0, z0], device=DEVICE)
#
#     # ------------------- x flip
#     if torch.rand(1, device=DEVICE) < FLIP_RATE:
#         image = image.flip(1)
#         masks = masks.flip(1)
#
#         if -1 not in skeletons:
#             for k, v in skeletons.items():
#                 skeletons[k][:, 0] = image.shape[1] - v[:, 0]
#
#     # ------------------- y flip
#     if torch.rand(1, device=DEVICE) < FLIP_RATE:
#         image = image.flip(2)
#         masks = masks.flip(2)
#         if -1 not in skeletons:
#             for k, v in skeletons.items():
#                 skeletons[k][:, 1] = image.shape[2] - v[:, 1]
#
#     # ------------------- z flip
#     if torch.rand(1, device=DEVICE) < FLIP_RATE:
#         image = image.flip(3)
#         masks = masks.flip(3)
#         if -1 not in skeletons:
#             for k, v in skeletons.items():
#                 skeletons[k][:, 2] = image.shape[3] - v[:, 2]
#
#     # # ------------------- Random Invert
#     if torch.rand(1, device=DEVICE) < BRIGHTNESS_RATE:
#         image = image.sub(1).mul(-1)
#
#     # ------------------- Adjust Brightness
#     if torch.rand(1, device=DEVICE) < BRIGHTNESS_RATE:
#         # funky looking but FAST
#         val = torch.empty(image.shape[0], device=DEVICE).uniform_(
#             BRIGHTNESS_RANGE[0], BRIGHTNESS_RANGE[1]
#         )
#         image = image.add(val.reshape(image.shape[0], 1, 1, 1)).clamp(0, 1)
#
#     # ------------------- Adjust Contrast
#     if torch.rand(1, device=DEVICE) < CONTRAST_RATE:
#         contrast_val = (CONTRAST_RANGE[1] - CONTRAST_RANGE[0]) * torch.rand(
#             (image.shape[0]), device=DEVICE
#         ) + CONTRAST_RANGE[0]
#
#         for z in range(image.shape[-1]):
#             image[..., z] = ttf.adjust_contrast(
#                 image[..., z], contrast_val[0].item()
#             ).squeeze(0)
#
#     # ------------------- Noise
#     if torch.rand(1, device=DEVICE) < NOISE_RATE:
#         noise = torch.rand(image.shape, device=DEVICE) * NOISE_GAMMA
#
#         image = image.add(noise).clamp(0, 1)
#
#     data_dict["image"] = image
#     data_dict["masks"] = masks
#     data_dict["skeletons"] = skeletons
#
#     baked: Tensor = bake_skeleton(
#         masks.squeeze(0).contigous(),
#         skeletons,
#         anisotropy=bake_skeleton_anisotropy,
#         average=True,
#         device=DEVICE,
#     )
#     data_dict["baked_skeleton"]: Union[Tensor, None] = baked
#
#     _, x, y, z = masks.shape
#     data_dict["skele_masks"]: Tensor = skeleton_to_mask(
#         skeletons, (x, y, z), kernel_size=smooth_skeleton_kernel_size, n=1
#     )
#
#     return data_dict
#
#
# def background_transform_3D(
#         data_dict: Dict[str, Tensor], device: Optional[str] = None
# ) -> DataDict:
#     DEVICE: str = str(data_dict["image"].device) if device is None else device
#
#     # Image should be in shape of [C, H, W, D]
#     CROP_WIDTH = torch.tensor([300], device=DEVICE)
#     CROP_HEIGHT = torch.tensor([300], device=DEVICE)
#     CROP_DEPTH = torch.tensor([20], device=DEVICE)
#
#     FLIP_RATE = torch.tensor(0.5, device=DEVICE)
#
#     BRIGHTNESS_RATE = torch.tensor(0.4, device=DEVICE)
#     INVERT_RATE = torch.tensor(0.5, device=DEVICE)
#     BRIGHTNESS_RANGE = torch.tensor((-0.1, 0.1), device=DEVICE)
#
#     NOISE_GAMMA = torch.tensor(0.1, device=DEVICE)
#     NOISE_RATE = torch.tensor(0.2, device=DEVICE)
#
#     FILTER_RATE = torch.tensor(0.5, device=DEVICE)
#
#     CONTRAST_RATE = torch.tensor(0.33, device=DEVICE)
#     CONTRAST_RANGE = torch.tensor((0.75, 2.0), device=DEVICE)
#
#     AFFINE_RATE = torch.tensor(0.66, device=DEVICE)
#     AFFINE_SCALE = torch.tensor((0.85, 1.1), device=DEVICE)
#     AFFINE_YAW = torch.tensor((-180, 180), device=DEVICE)
#     AFFINE_SHEAR = torch.tensor((-7, 7), device=DEVICE)
#
#     image = torch.clone(data_dict["image"])  #
#
#     # ------------ Random Crop 1
#     extra = 300
#     w = (
#         CROP_WIDTH + extra
#         if CROP_WIDTH + extra <= image.shape[1]
#         else torch.tensor(image.shape[1])
#     )
#     h = (
#         CROP_HEIGHT + extra
#         if CROP_HEIGHT + extra <= image.shape[2]
#         else torch.tensor(image.shape[2])
#     )
#     d = CROP_DEPTH if CROP_DEPTH <= image.shape[3] else torch.tensor(image.shape[3])
#
#     shape = torch.tensor(image.shape[1::], device=DEVICE) - torch.tensor(
#         [w, h, d], device=DEVICE
#     )
#     center = torch.tensor(
#         [torch.randint(0, s, (1,)).item() if s > 0 else 0 for s in shape], device=DEVICE
#     )
#
#     # Center that instance
#     x0 = (
#         center[0]
#         .sub(torch.floor(w / 2))
#         .long()
#         .clamp(min=0, max=image.shape[1] - w.item())
#     )
#     y0 = (
#         center[1]
#         .sub(torch.floor(h / 2))
#         .long()
#         .clamp(min=0, max=image.shape[2] - h.item())
#     )
#     z0 = (
#         center[2]
#         .sub(torch.floor(d / 2))
#         .long()
#         .clamp(min=0, max=image.shape[3] - d.item())
#     )
#
#     x1 = x0 + w
#     y1 = y0 + h
#     z1 = z0 + d
#
#     image = image[
#             :, x0.item(): x1.item(), y0.item(): y1.item(), z0.item(): z1.item()
#             ].to(DEVICE)
#
#     # -------------------affine (Cant use baked skeletons)
#     if torch.rand(1, device=DEVICE) < AFFINE_RATE:
#         angle = (AFFINE_YAW[1] - AFFINE_YAW[0]) * torch.rand(
#             1, device=DEVICE
#         ) + AFFINE_YAW[0]
#         shear = (AFFINE_SHEAR[1] - AFFINE_SHEAR[0]) * torch.rand(
#             1, device=DEVICE
#         ) + AFFINE_SHEAR[0]
#         scale = (AFFINE_SCALE[1] - AFFINE_SCALE[0]) * torch.rand(
#             1, device=DEVICE
#         ) + AFFINE_SCALE[0]
#
#         image = ttf.affine(
#             image.permute(0, 3, 1, 2).float(),
#             angle=angle.item(),
#             shear=[float(shear.item())],
#             scale=scale.item(),
#             translate=[0, 0],
#         ).permute(0, 2, 3, 1)
#
#     # ------------ Center Crop 2
#     w = CROP_WIDTH if CROP_WIDTH <= image.shape[1] else torch.tensor(image.shape[1])
#     h = CROP_HEIGHT if CROP_HEIGHT <= image.shape[2] else torch.tensor(image.shape[2])
#     d = CROP_DEPTH if CROP_DEPTH <= image.shape[3] else torch.tensor(image.shape[3])
#
#     center = center - torch.tensor([x0, y0, z0], device=DEVICE)
#
#     # Center that instance
#     x0 = (
#         center[0]
#         .sub(torch.floor(w / 2))
#         .long()
#         .clamp(min=0, max=image.shape[1] - w.item())
#     )
#     y0 = (
#         center[1]
#         .sub(torch.floor(h / 2))
#         .long()
#         .clamp(min=0, max=image.shape[2] - h.item())
#     )
#     z0 = (
#         center[2]
#         .sub(torch.floor(d / 2))
#         .long()
#         .clamp(min=0, max=image.shape[3] - d.item())
#     )
#
#     x1 = x0 + w
#     y1 = y0 + h
#     z1 = z0 + d
#
#     image = image[
#             :, x0.item(): x1.item(), y0.item(): y1.item(), z0.item(): z1.item()
#             ]
#
#     # ------------------- x flip
#     if torch.rand(1, device=DEVICE) < FLIP_RATE:
#         image = image.flip(1)
#
#     # ------------------- y flip
#     if torch.rand(1, device=DEVICE) < FLIP_RATE:
#         image = image.flip(2)
#
#     # ------------------- z flip
#     if torch.rand(1, device=DEVICE) < FLIP_RATE:
#         image = image.flip(3)
#
#     # # ------------------- Random Invert
#     if torch.rand(1, device=DEVICE) < INVERT_RATE:
#         image = image.sub(1).mul(-1)
#
#     # ------------------- Adjust Brightness
#     if torch.rand(1, device=DEVICE) < BRIGHTNESS_RATE:
#         # funky looking but FAST
#         val = torch.empty(image.shape[0], device=DEVICE).uniform_(
#             BRIGHTNESS_RANGE[0], BRIGHTNESS_RANGE[1]
#         )
#         image = image.add(val.reshape(image.shape[0], 1, 1, 1)).clamp(0, 1)
#
#     # ------------------- Adjust Contrast
#     if torch.rand(1, device=DEVICE) < CONTRAST_RATE:
#         contrast_val = (CONTRAST_RANGE[1] - CONTRAST_RANGE[0]) * torch.rand(
#             (image.shape[0]), device=DEVICE
#         ) + CONTRAST_RANGE[0]
#
#         for z in range(image.shape[-1]):
#             image[..., z] = ttf.adjust_contrast(
#                 image[..., z], contrast_val[0].item()
#             ).squeeze(0)
#
#     # ------------------- Noise
#     if torch.rand(1, device=DEVICE) < NOISE_RATE:
#         noise = torch.rand(image.shape, device=DEVICE) * NOISE_GAMMA
#
#         image = image.add(noise).clamp(0, 1)
#
#     data_dict["image"] = image
#     data_dict["masks"] = torch.zeros_like(image, device=DEVICE)
#     data_dict["skeletons"]: Dict[int, Tensor] = {-1: torch.empty((0, 3), device=DEVICE)}
#     data_dict["baked_skeleton"] = torch.zeros(
#         (3, image.shape[1], image.shape[2], image.shape[3]), device=DEVICE
#     )
#     data_dict["skele_masks"] = torch.zeros_like(image, device=DEVICE)
#
#     return data_dict


if __name__ == "__main__":
    import torch.distributed as dist
    import torch.optim.lr_scheduler

    import skimage.io as io
    import matplotlib.pyplot as plt
    import skoots.train.generate_skeletons
    from skoots.lib.skeleton import skeleton_to_mask

    img = torch.from_numpy(
        io.imread(
            "/home/chris/Dropbox (Partners HealthCare)/skoots/tests/test_data/hide_validate_skeleton_instance_mask.tif"
        ).astype(float)
    )

    mask = torch.from_numpy(
        io.imread(
            "/home/chris/Dropbox (Partners HealthCare)/skoots/tests/test_data/hide_validate.labels.tif"
        ).astype(float)
    )

    img = img[:, 850:1100, 690:890].permute(1, 2, 0)
    mask = mask[:, 850:1100, 690:890].permute(1, 2, 0)

    skeletons = skoots.train.generate_skeletons.calculate_skeletons(mask, torch.tensor((0.3, 0.3, 1.0)))

    x, y, z = img.shape

    img = img.view(1, 1, x, y, z)
    mask = mask.view(1, 1, x, y, z)

    og = img.clone()

    og_skl = skeleton_to_mask(skeletons, (x, y, z))

    img, mask, skeletons = elastic_deform(img, mask, skeletons)

    skl_msk = skeleton_to_mask(skeletons, (x, y, z))

    plt.imshow(og[0, 0, :, :, 14])
    plt.show()

    plt.imshow(og_skl[0, :, :, 14])
    plt.show()

    plt.imshow(img[0, 0, :, :, 14])
    plt.show()

    plt.imshow(skl_msk[0, :, :, 14])
    plt.show()



from typing import Dict, List

import torch
from yacs.config import CfgNode


# DOCUMENTED


class Sigma:
    """an adjustable sigma parameter for training."""

    def __init__(
        self,
        adjustments: List[Dict[str, float]],
        initial_sigma: List[float] = [0.1, 0.1, 0.8],
        device="cpu",
    ):
        """
        Creates an object which inputs an epoch, and returns a torch.tensor of [sigma_x, sigma_y, sigma_z]

        Sigma should reflect the Error of embedding vectors at each spatial dim.

        :param adjustments: lost of adjustments and when to apply them
        :param initial_sigma: initial values of sigma at epoch=0
        :param device: device to load sigma on ('cpu' or 'cuda')
        """
        self.adjutments = adjustments
        self.device = device

        self.initial_sigma = (
            torch.tensor(initial_sigma, device=device)
            if isinstance(initial_sigma, list)
            else initial_sigma
        )

        values = [1]  # Initial values set so sigma isnt zero at 3 zero
        epochs = [-1]
        for d in adjustments:
            values.append(d["multiplier"])
            epochs.append(d["epoch"])

        self.values = torch.tensor(values, device=self.device)
        self.epochs = torch.tensor(epochs, device=self.device)

    def __call__(self, e: int):
        """
        Returns sigma at epoch "e"
        :param e: epoch
        :return: sigma [x,y,z]
        """
        multiplier = self.values[self.epochs < e].prod()

        return self.initial_sigma * multiplier


def init_sigma(cfg: CfgNode, device: torch.device) -> Sigma:
    start = torch.tensor(cfg.TRAIN.INITIAL_SIGMA, device=device)
    multipliers = [{"multiplier": a, "epoch": b} for a, b in cfg.TRAIN.SIGMA_DECAY]
    return Sigma(initial_sigma=start, adjustments=multipliers)


if __name__ == "__main__":
    from skoots.config import get_cfg_defaults

    cfg = get_cfg_defaults()

    _ = init_sigma(cfg, "cpu")



import skoots.train.loss
from typing import Dict, Tuple

import torch
from torch import Tensor
from tqdm import tqdm

import skoots.train.loss
from skoots.validate.utils import imread


def mask_to_bbox(mask: Tensor) -> Tuple[Tensor, Tensor]:
    """
    Calculates the 3D bbox for each instance of an instance segmentation mask.

    Assumes each positive integer is an instance for class label. Returns a tensor of id labels and a tensor of bboxes
    bboxes are in format: [x0,y0,z0,x1,y1,z1]

    Assigns a bbox to each unique lablel! Does not mean each label has a valid bbox!!!

    Shapes:
        - mask: :math:`(1, X_{in}, Y_{in}, Z_{in})`
        - return[0]: Id labels: :math: `(N)`
        - return[1]: bboxes: :math: `(6, N)`


    :param mask: Input instance segmentation mask
    :return: id labels and bboxes
    """
    assert mask.ndim == 4, "Mask ndim != 4"

    sparse_mask = mask.to_sparse()

    indices = sparse_mask.indices()
    values = sparse_mask.values()

    unique = torch.unique(values)  # assured to have no zeros because tensor is sparse

    bboxes = torch.empty(
        (6, unique.shape[0]), device=mask.device, dtype=torch.int16
    )  # preallocate for speed

    for i, u in enumerate(unique):
        ind = indices[1::, values == u]  # just x,y,z dim of indicies

        bboxes[0, i] = ind[0].min()  # x0
        bboxes[1, i] = ind[1].min()  # y0
        bboxes[2, i] = ind[2].min()  # z0

        bboxes[3, i] = ind[0].max()  # x1
        bboxes[4, i] = ind[1].max()  # y1
        bboxes[5, i] = ind[2].max()  # z1

    return unique, bboxes


def valid_box_inds(boxes):
    """
    returns the inds of all valid boxes

    :param boxes: [6, N]
    :return: [N]
    """

    x0, y0, z0, x1, y1, z1 = (
        boxes[0, :],
        boxes[1, :],
        boxes[2, :],
        boxes[3, :],
        boxes[4, :],
        boxes[5, :],
    )
    inds = torch.logical_and(torch.logical_and((x1 > x0), (y1 > y0)), (z1 > z0))
    return inds


def box_iou(a: Tensor, b: Tensor) -> Tensor:
    """
    Compute the IoU of the cartesian product of two sets of boxes.

    Each box in each set shall be (x0, y0, z0, x0, y0, z0).

    Shapes:
        a: :math:`(6, N)`.
        b: :math:`(6, M)`.
        returns: :math: `(N, M)`


    :param a: box 1
    :param b: box 2
    :return: iou of each box in 1 against all boxes in 2
    """
    if not valid_box_inds(a).bool().all():
        raise AssertionError("a does not follow (x0, y0, z0, x1, y1, z1) format.")

    if not valid_box_inds(b).bool().all():
        raise AssertionError("b does not follow (x0, y0, z0, x1, y2, z1) format.")

    a = a.T.float()  # we'll have buffer overflow otherwise
    b = b.T.float()  # we'll have buffer overflow otherwise

    # find intersection
    lower_bounds = torch.max(
        a[:, :3].unsqueeze(1), b[:, :3].unsqueeze(0)
    ).float()  # (n, m, 3)
    upper_bounds = torch.min(
        a[:, 3:].unsqueeze(1), b[:, 3:].unsqueeze(0)
    ).float()  # (n, m, 3)

    intersection_dims = torch.clamp(upper_bounds - lower_bounds, min=0)  # (n1, n2, 3)

    intersection = (
        intersection_dims[:, :, 0].float()
        * intersection_dims[:, :, 1].float()
        * intersection_dims[:, :, 2].float()
    )  # (n, m)

    # Find areas of each box in both sets
    areas_a = (a[:, 3] - a[:, 0]) * (a[:, 4] - a[:, 1]) * (a[:, 5] - a[:, 2])  # (n)

    areas_b = (b[:, 3] - b[:, 0]) * (b[:, 4] - b[:, 1]) * (b[:, 5] - b[:, 2])  # (m)

    union = areas_a.unsqueeze(1) + areas_b.unsqueeze(0) - intersection  # (n1, n2)

    return intersection / union  # (n1, n2)


def calculate_accuracies_from_bbox(
    ground_truth: Dict[str, Tensor],
    predictions: Dict[str, Tensor],
    device: str | None = None,
    threshold=0.1,
):
    """
    Calculates True positive, False Positive, False Negative from data_dict of segmentation 3d bboxes

    :param ground_truth:
    :param predictions:
    :param device:
    :param threshold:
    :return:
    """

    device = device if device else ground_truth["boxes"].device

    _gt = ground_truth["boxes"].to(device)
    _pred = predictions["boxes"].to(device)

    iou = box_iou(_gt, _pred)

    gt_max, gt_indicies = iou.max(dim=1)
    gt = torch.logical_not(gt_max.gt(threshold)) if iou.shape[1] > 0 else torch.ones(0)
    pred = (
        torch.logical_not(iou.max(dim=0)[0].gt(threshold))
        if iou.shape[0] > 0
        else torch.ones(0)
    )

    true_positive = torch.sum(torch.logical_not(gt))
    false_positive = torch.sum(pred)
    false_negative = torch.sum(gt)

    return (
        true_positive,
        false_positive,
        false_negative,
    )


def accuracies_from_iou(iou: Tensor, thr: float = 0.1) -> Tensor:
    gt_max, gt_indicies = iou.max(dim=1)
    gt = torch.logical_not(gt_max.gt(thr)) if iou.shape[1] > 0 else torch.ones(0)
    pred = (
        torch.logical_not(iou.max(dim=0)[0].gt(thr))
        if iou.shape[0] > 0
        else torch.ones(0)
    )

    true_positive = torch.sum(torch.logical_not(gt))
    false_positive = torch.sum(pred)
    false_negative = torch.sum(gt)

    return (
        true_positive.cpu().item(),
        false_positive.cpu().item(),
        false_negative.cpu().item(),
    )


def mask_iou(gt: Tensor, pred: Tensor):
    """
    Calculates the IoU of each object on a per-mask-basis.

    :param gt: mask 1 with N instances
    :param pred: mask 2 with M instances
    :return: NxM matrix of IoU's
    """
    assert gt.shape == pred.shape, "Input tensors must be the same shape"
    assert gt.device == pred.device, "Input tensors must be on the same device"

    a_unique = gt.unique()
    a_unique = a_unique[a_unique > 0]

    b_unique = pred.unique()
    b_unique = b_unique[b_unique > 0]

    iou = torch.zeros(
        (a_unique.shape[0], b_unique.shape[0]), dtype=torch.float, device=gt.device
    )

    for i, au in tqdm(enumerate(a_unique), total=len(a_unique)):
        _a = gt == au
        touching = pred[
            _a
        ].unique()  # we only calculate iou of lables which have "contact with" our mask
        touching = touching[touching != 0]

        for j, bu in enumerate(b_unique):
            if torch.any(touching == bu):
                _b = pred == bu

                intersection = torch.logical_and(_a, _b).sum()
                union = torch.logical_or(_a, _b).sum()

                iou[i, j] = intersection / union
            else:
                iou[i, j] = 0.0

    return iou


def mask_dice(gt: Tensor, pred: Tensor):
    """
    Calculates the Dice Index of each object on a per-mask-basis.

    :param gt: mask 1 with N instances
    :param pred: mask 2 with M instances
    :return: NxM matrix of IoU's
    """
    assert gt.shape == pred.shape, "Input tensors must be the same shape"
    assert gt.device == pred.device, "Input tensors must be on the same device"

    a_unique = gt.unique()
    a_unique = a_unique[a_unique > 0]

    b_unique = pred.unique()
    b_unique = b_unique[b_unique > 0]

    dice = torch.zeros(
        (a_unique.shape[0], b_unique.shape[0]), dtype=torch.float, device=gt.device
    )

    for i, au in tqdm(enumerate(a_unique), total=len(a_unique)):
        _a = gt == au
        touching = pred[
            _a
        ].unique()  # we only calculate iou of lables which have "contact with" our mask
        touching = touching[touching != 0]

        for j, bu in enumerate(b_unique):
            if torch.any(touching == bu):
                _b = pred == bu

                numerator = torch.logical_and(_a, _b).sum() * 2
                denominator = _a.sum() + _b.sum()

                assert (
                    numerator < denominator
                ), f"{numerator=}, {denominator=}, {_a.sum()=}, {_b.sum()=}, {(_a*_b).sum()=}"

                dice[i, j] = numerator / denominator
            else:
                dice[i, j] = 0.0

    return dice


def mask_soft_cldice(gt: Tensor, pred: Tensor):
    """
    Calculates the Dice Index of each object on a per-mask-basis.

    :param gt: mask 1 with N instances
    :param pred: mask 2 with M instances
    :return: NxM matrix of IoU's
    """
    assert gt.shape == pred.shape, "Input tensors must be the same shape"
    assert gt.device == pred.device, "Input tensors must be on the same device"

    a_unique = gt.unique()
    a_unique = a_unique[a_unique > 0]

    b_unique = pred.unique()
    b_unique = b_unique[b_unique > 0]

    criterion = torch.compile(skoots.train.loss.soft_cldice())
    cldice = torch.zeros(
        (a_unique.shape[0], b_unique.shape[0]), dtype=torch.float, device=gt.device
    )

    for i, au in tqdm(enumerate(a_unique), total=len(a_unique)):
        _a = gt == au
        touching = pred[
            _a
        ].unique()  # we only calculate iou of lables which have "contact with" our mask
        touching = touching[touching != 0]

        for j, bu in enumerate(b_unique):
            if torch.any(touching == bu):
                _b = pred == bu
                cldice[i, j] = criterion(_b.float(), _a.float())
            else:
                cldice[i, j] = 0.0

    return cldice


def sparse_mask_iou(a: Tensor, b: Tensor) -> Tensor:
    """
    Calculates the IoU of each object on a per-mask-basis using sparse tensors.

    :param a: mask 1 with N instances
    :param b: mask 2 with M instances
    :return: NxM matrix of IoU's
    """
    raise NotImplementedError("In Development...")

    assert a.shape == b.shape, "Input tensors must be the same shape"
    assert a.device == b.device, "Input tensors must be on the same device"

    shape = a.shape

    a = a.to_sparse_coo()
    b = b.to_sparse_coo()

    a_unique = a.labels().unique()
    b_unique = b.labels().unique()

    a_indicies = a.indicies()
    b_indicies = b.indicies()

    iou = torch.zeros(
        (a_unique.shape[0], b_unique.shape[0]), dtype=torch.float, device=a.device
    )

    for i, au in tqdm(enumerate(a_unique)):
        for j, bu in enumerate(b_unique):
            _a_ind = a_indicies[a.lables() == au]
            _b_ind = b_indicies[b.lables() == bu]

            a_sparse = torch.sparse_coo_tensor(
                indices=_a_ind, labels=torch.ones_like(a.labels() == au), size=shape
            )
            b_sparse = torch.sparse_coo_tensor(
                indices=_b_ind, labels=torch.ones_like(b.labels() == bu), size=shape
            )


def f1_score(tp, fp, fn):
    num = 2 * tp
    dem = 2 * tp + fp + fn
    return num / dem


def _iou_instance_dict(a: Tensor, b: Tensor) -> Dict[int, Tensor]:
    """
    Given two instance masks, compares each instance in b against a. Usually assumes A is the ground truth.

    :param a: Mask A
    :param b: Mask B
    :return:  Dict of instances and every IOU for each instance
    """
    a_unique = a.unique()
    a_unique = a_unique[a_unique > 0]

    b_unique = b.unique()
    b_unique = b_unique[b_unique > 0]

    iou = {}

    for i, au in tqdm(enumerate(a_unique), total=len(a_unique)):
        _a = a == au

        touching = b[
            _a
        ].unique()  # we only calculate iou of lables which have "contact with" our mask
        touching = touching[touching != 0]
        iou[au] = []

        for j, bu in enumerate(b_unique):
            if torch.any(touching == bu):
                _b = b == bu

                intersection = torch.logical_and(_a, _b).sum()
                union = torch.logical_or(_a, _b).sum()
                iou[au].append((intersection / union).item())

    return iou


def get_segmentation_errors(ground_truth: Tensor, predicted: Tensor) -> float:
    """
    Calculates the IoU of each object on a per-mask-basis.

    :param ground_truth: mask 1 with N instances
    :param predicted: mask 2 with M instances
    :return: NxM matrix of IoU's
    """

    iou = _iou_instance_dict(ground_truth, predicted)
    for k, v in iou.items():
        iou[k] = torch.tensor(v)

    num_split = 0
    for k, v in iou.items():
        if v.gt(0.2).int().sum() > 1:
            num_split += 1

    over_segmentation_rate = num_split / len(iou)

    iou = _iou_instance_dict(predicted, ground_truth)
    for k, v in iou.items():
        iou[k] = torch.tensor(v)

    num_split = 0
    for k, v in iou.items():
        if v.gt(0.2).int().sum() > 1:
            num_split += 1

    under_segmentation_rate = num_split / len(iou)

    return over_segmentation_rate, under_segmentation_rate


if __name__ == "__main__":
    gt = imread("../../tests/test_data/hide_validate.labels.tif")[..., 2:-2]
    pred = imread("../../tests/test_data/hide_validate_skeleton_instance_mask.tif")[
        ..., 2:-2
    ]
    aff_pred = imread(
        "../../tests/test_data/hide_validation_affinity_instance_segmentaiton.tif"
    )[..., 2:-2]

    device = "cuda" if torch.cuda.is_available() else "cpu"
    gt = gt.to(device)
    pred = pred.to(device)
    aff_pred = aff_pred.to(device)

    u, c = aff_pred.unique(return_counts=True)
    for a, b in tqdm(zip(u, c)):
        if b < 500:
            aff_pred[aff_pred == a] = 0

    skoots_seg_errors = get_segmentation_errors(gt, pred)
    aff_seg_errors = get_segmentation_errors(gt, aff_pred)

    # print('SKOOTS IOU')
    # if not os.path.exists('../../tests/iou_gt_skoots_mask.trch'):
    #     iou_skoots = mask_iou(gt, pred)
    #     torch.save(iou_skoots, '../../tests/iou_gt_skoots_mask.trch')
    # else:
    #     iou_skoots = torch.load('../../tests/iou_gt_skoots_mask.trch')
    #
    # print('AFFINITES IOU')
    # if not os.path.exists('../../tests/iou_gt_affinites_mask.trch'):
    #     iou_aff = mask_iou(gt, aff_pred)
    #     torch.save(iou_aff, '../../tests/iou_gt_affinites_mask.trch')
    # else:
    #     iou_aff = torch.load('../../tests/iou_gt_affinites_mask.trch')
    #
    # # iou_skoots = torch.load('../../tests/iou_gt_skoots_mask.trch')
    # # iou_aff = torch.load('../../tests/iou_gt_affinites_mask.trch')
    #
    #
    #
    # tfp_skoots = [accuracies_from_iou(iou_skoots, thr/100) for thr in range(100)]
    # tfp_aff = [accuracies_from_iou(iou_aff, thr/100) for thr in range(100)]
    #
    # precision_skoots = [(tp /(tp + fp)) for (tp, fp, fn) in tfp_skoots]
    # recall_skoots = [(tp / (tp + fn)) for (tp, fp, fn) in tfp_skoots]
    #
    # precision_aff = [(tp/(tp+fp)) for (tp, fp, fn) in tfp_aff]
    # recall_aff = [(tp / (tp + fn)) for (tp, fp, fn) in tfp_aff]
    #
    # f1_skoots = [f1_score(*a) for a in tfp_skoots]
    # f1_aff = [f1_score(*a) for a in tfp_aff]
    #
    # plt.plot(np.arange(0, 100), precision_skoots)
    # plt.plot(np.arange(0, 100), precision_aff)
    # plt.legend(['SKOOTS', 'AFFINITIES'])
    # plt.ylabel('Precision')
    # plt.xlabel('IoU Threshold: (%)')
    # plt.show()
    #
    # plt.plot(np.arange(0, 100), recall_skoots)
    # plt.plot(np.arange(0, 100), recall_aff)
    # plt.legend(['SKOOTS', 'AFFINITIES'])
    # plt.ylabel('Recall')
    # plt.xlabel('IoU Threshold: (%)')
    # plt.show()
    #
    # plt.plot(np.arange(0, 100), f1_skoots)
    # plt.plot(np.arange(0, 100), f1_aff)
    # plt.legend(['SKOOTS', 'AFFINITIES'])
    # plt.ylabel('F1')
    # plt.xlabel('IoU Threshold: (%)')
    # plt.show()



import os
import os.path
from functools import partial
from statistics import mean
from typing import Callable, Union, Dict

import skoots.train.loss
import torch
import torch.nn as nn
import torch.optim.lr_scheduler
import torch.optim.swa_utils
from skoots.lib.embedding_to_prob import baked_embed_to_prob
from skoots.lib.vector_to_embedding import vector_to_embedding
from skoots.train.dataloader import dataset, MultiDataset, skeleton_colate
from skoots.train.merged_transform import (
    TransformFromCfg,
    BackgroundTransformFromCfg,
)
from skoots.train.setup import setup_process
from skoots.train.sigma import Sigma, init_sigma
from skoots.train.utils import write_progress
from torch import Tensor
from torch import profiler
from torch.cuda.amp import GradScaler, autocast
from torch.utils.data import DataLoader, Dataset
from torch.utils.tensorboard import SummaryWriter
from tqdm import trange
from yacs.config import CfgNode

import logging

Dataset = Union[Dataset, DataLoader]

_valid_optimizers = {
    "adamw": torch.optim.AdamW,
    "adam": torch.optim.Adam,
    "sgd": torch.optim.SGD,
    "adamax": torch.optim.Adamax,
}

_valid_loss_functions = {
    "soft_cldice": skoots.train.loss.soft_dice_cldice,
    "tversky": skoots.train.loss.tversky,
}

_valid_lr_schedulers = {
    "cosine_annealing_warm_restarts": torch.optim.lr_scheduler.CosineAnnealingWarmRestarts
}

torch.manual_seed(101196)
torch.set_float32_matmul_precision("high")


def train(rank: str, port: str, world_size: int, base_model: nn.Module, cfg: CfgNode, logging_level: int):
    setup_process(rank, world_size, port, backend="nccl")
    device = f"cuda:{rank}"

    _log_map = [
        logging.DEBUG,
        logging.INFO,
        logging.WARNING,
        logging.ERROR,
        logging.CRITICAL,
    ]
    logging.basicConfig(
        level=_log_map[logging_level],
        format=f"[%(asctime)s] skoots-train rank{rank} [%(levelname)s]: %(message)s",
        force=True,
    )

    base_model = base_model.to(device)
    base_model = torch.nn.parallel.DistributedDataParallel(base_model)

    if int(rank) == 0:
        print(cfg)

    if int(torch.__version__[0]) >= 2:
        logging.info('Compiling model with torch inductor')
        model = torch.compile(base_model)
    model=base_model

    augmentations = TransformFromCfg(cfg=cfg, device=device).set_dataset_std(255).set_dataset_mean(0)
    background_augmentations = BackgroundTransformFromCfg(cfg=cfg, device=device).set_dataset_std(255).set_dataset_mean(0)


    assert len(cfg.TRAIN.TRAIN_STORE_DATA_ON_GPU) == len(
        cfg.TRAIN.TRAIN_DATA_DIR), 'must set identical length gpu storage argument'
    assert len(cfg.TRAIN.VALIDATION_STORE_DATA_ON_GPU) == len(
        cfg.TRAIN.VALIDATION_DATA_DIR), 'must set identical length gpu storage argument'
    assert len(cfg.TRAIN.BACKGROUND_STORE_DATA_ON_GPU) == len(
        cfg.TRAIN.BACKGROUND_DATA_DIR), 'must set identical length gpu storage argument'

    _datasets = []
    for path, N, on_cuda in zip(cfg.TRAIN.TRAIN_DATA_DIR, cfg.TRAIN.TRAIN_SAMPLE_PER_IMAGE,
                                cfg.TRAIN.TRAIN_STORE_DATA_ON_GPU):
        _device = device if on_cuda else 'cpu'
        _datasets.append(
            dataset(
                path=path,
                transforms=augmentations,
                sample_per_image=N,
                device=device,
                pad_size=10,
            )
            .pin_memory()
            .to(_device)
        )

    for path, N, on_cuda in zip(
            cfg.TRAIN.BACKGROUND_DATA_DIR, cfg.TRAIN.BACKGROUND_SAMPLE_PER_IMAGE, cfg.TRAIN.BACKGROUND_STORE_DATA_ON_GPU
    ):
        _device = device if on_cuda else 'cpu'
        _datasets.append(
            dataset(
                path=path,
                transforms=background_augmentations,
                sample_per_image=N,
                device=device,
                pad_size=100,
            )
            .pin_memory()
            .to(_device)
        )

    merged_train = MultiDataset(*_datasets)

    train_sampler = torch.utils.data.distributed.DistributedSampler(merged_train)
    _n_workers = 0  # if _device != 'cpu' else 2
    dataloader = DataLoader(
        merged_train,
        num_workers=_n_workers,
        batch_size=cfg.TRAIN.TRAIN_BATCH_SIZE,
        sampler=train_sampler,
        collate_fn=skeleton_colate,
    )

    # Validation Dataset
    _datasets = []
    for path, N, on_cuda in zip(
            cfg.TRAIN.VALIDATION_DATA_DIR, cfg.TRAIN.VALIDATION_SAMPLE_PER_IMAGE, cfg.TRAIN.VALIDATION_STORE_DATA_ON_GPU
    ):
        _device = device if on_cuda else 'cpu'
        _datasets.append(
            dataset(
                path=path,
                transforms=augmentations,
                sample_per_image=N,
                device=device,
                pad_size=10,
            )
            .pin_memory()
            .to(_device)
        )

    merged_validation = MultiDataset(*_datasets)
    test_sampler = torch.utils.data.distributed.DistributedSampler(merged_validation)
    if _datasets or cfg.TRAIN.VALIDATION_BATCH_SIZE >= 1:
        _n_workers = 0  # if _device != 'cpu' else 2
        _device = device if cfg.TRAIN.STORE_DATA_ON_GPU else "cpu"
        valdiation_dataloader = DataLoader(
            merged_validation,
            num_workers=_n_workers,
            batch_size=cfg.TRAIN.VALIDATION_BATCH_SIZE,
            sampler=test_sampler,
            collate_fn=skeleton_colate,
        )

    else:  # we might not want to run validation...
        valdiation_dataloader = None

    torch.backends.cudnn.benchmark = cfg.TRAIN.CUDNN_BENCHMARK
    torch.autograd.profiler.profile = cfg.TRAIN.AUTOGRAD_PROFILE
    torch.autograd.profiler.emit_nvtx(enabled=cfg.TRAIN.AUTOGRAD_EMIT_NVTX)
    torch.autograd.set_detect_anomaly(cfg.TRAIN.AUTOGRAD_DETECT_ANOMALY)

    sigma: Sigma = init_sigma(cfg, device)
    epochs = cfg.TRAIN.NUM_EPOCHS

    writer = SummaryWriter() if rank == 0 else None
    if writer:
        print("SUMMARY WRITER LOG DIR: ", writer.get_logdir())

    # TRAIN LOOP ----------------------------

    vector_scale = torch.tensor(cfg.SKOOTS.VECTOR_SCALING, device=device)

    kwargs = {k:v for k, v in zip(cfg.TRAIN.OPTIMIZER_KEYWORD_ARGUMENTS, cfg.TRAIN.OPTIMIZER_KEYWORD_VALUES)}
    logging.info('creating optimizer')
    optimizer = _valid_optimizers[cfg.TRAIN.OPTIMIZER](
        model.parameters(),
        lr=cfg.TRAIN.LEARNING_RATE,
        weight_decay=cfg.TRAIN.WEIGHT_DECAY, **kwargs
    )

    logging.info('creating scheduler')
    scheduler = _valid_lr_schedulers[cfg.TRAIN.SCHEDULER](
        optimizer, T_0=cfg.TRAIN.SCHEDULER_T0
    )
    logging.info('creating grad scaler')
    scaler = GradScaler(enabled=cfg.TRAIN.MIXED_PRECISION)

    swa_model = torch.optim.swa_utils.AveragedModel(model)
    swa_start = 100
    swa_scheduler = torch.optim.swa_utils.SWALR(optimizer, swa_lr=0.05)

    logging.info('creating loss functions')
    _kwarg = {
        k: v for k, v in zip(cfg.TRAIN.LOSS_EMBED_KEYWORDS, cfg.TRAIN.LOSS_EMBED_VALUES)
    }
    loss_embed: Callable = _valid_loss_functions[cfg.TRAIN.LOSS_EMBED](**_kwarg)

    _kwarg = {
        k: v
        for k, v in zip(
            cfg.TRAIN.LOSS_PROBABILITY_KEYWORDS, cfg.TRAIN.LOSS_PROBABILITY_VALUES
        )
    }
    loss_prob: Callable = _valid_loss_functions[cfg.TRAIN.LOSS_PROBABILITY](**_kwarg)

    _kwarg = {
        k: v
        for k, v in zip(
            cfg.TRAIN.LOSS_SKELETON_KEYWORDS, cfg.TRAIN.LOSS_SKELETON_VALUES
        )
    }
    loss_skele: Callable = _valid_loss_functions[cfg.TRAIN.LOSS_SKELETON](**_kwarg)

    # Save each loss value in a list...
    avg_epoch_loss = [9999999999.9999999999]
    avg_epoch_embed_loss = [9999999999.9999999999]
    avg_epoch_prob_loss = [9999999999.9999999999]
    avg_epoch_skele_loss = [9999999999.9999999999]

    avg_val_loss = [9999999999.9999999999]
    avg_val_embed_loss = [9999999999.9999999999]
    avg_val_prob_loss = [9999999999.9999999999]
    avg_val_skele_loss = [9999999999.9999999999]

    # skel_crossover_loss = skoots.train.loss.split(n_iter=3, alpha=2)

    # Warmup... Get the first from train_data
    logging.info('generating warmup data')
    for images, masks, skeleton, skele_masks, baked in dataloader:
        break

    assert images is not None, len(dataloader)

    logging.info('starting warmup process')
    warmup_range = trange(cfg.TRAIN.N_WARMUP, desc="Warmup: {}")
    for w in warmup_range:
        optimizer.zero_grad(set_to_none=True)

        with autocast(enabled=cfg.TRAIN.MIXED_PRECISION):  # Saves Memory!
            out: Tensor = model(images)

            probability_map: Tensor = out[:, [-1], ...]
            vector: Tensor = out[:, 0:3:1, ...]
            predicted_skeleton: Tensor = out[:, [-2], ...]

            embedding: Tensor = vector_to_embedding(vector_scale, vector)
            out: Tensor = baked_embed_to_prob(embedding, baked, sigma(0))

            _loss_embed = loss_embed(
                out, masks.gt(0).float()
            )  # out = [B, 2/3, X, Y, Z?]
            _loss_prob = loss_prob(probability_map, masks.gt(0).float())
            _loss_skeleton = loss_skele(
                predicted_skeleton, skele_masks.gt(0).float()
            )  # + skel_crossover_loss(predicted_skeleton, skele_masks.gt(0).float())

            loss = (
                    (cfg.TRAIN.LOSS_EMBED_RELATIVE_WEIGHT * _loss_embed)
                    + (cfg.TRAIN.LOSS_PROBABILITY_RELATIVE_WEIGHT * _loss_prob)
                    + (cfg.TRAIN.LOSS_SKELETON_RELATIVE_WEIGHT * _loss_skeleton)
            )

            warmup_range.desc = f"{loss.item()}"

            if torch.isnan(loss):
                print(
                    f"Found NaN value in loss.\n\tLoss Embed: {_loss_embed}\n\tLoss Probability: {_loss_prob}"
                )
                print(f"\t{torch.any(torch.isnan(vector))}")
                print(f"\t{torch.any(torch.isnan(embedding))}")
                continue

        scaler.scale(loss).backward()
        scaler.step(optimizer)
        scaler.update()

    # Train Step...
    epoch_range = (
        trange(epochs, desc=f"Loss = {1.0000000}") if rank == 0 else range(epochs)
    )
    for e in epoch_range:
        logging.info(f'Starting Epoch: {e+1}/{epochs}')
        _loss, _embed, _prob, _skele = [], [], [], []

        if cfg.TRAIN.DISTRIBUTED:
            train_sampler.set_epoch(e)

        for index, (images, masks, skeleton, skele_masks, baked) in enumerate(dataloader):
            optimizer.zero_grad(set_to_none=True)
            logging.info(f'Starting Batch: {index+1}/{len(dataloader)}')

            with autocast(enabled=cfg.TRAIN.MIXED_PRECISION):  # Saves Memory!
                out: Tensor = model(images)

                probability_map: Tensor = out[:, [-1], ...]
                vector: Tensor = out[:, 0:3:1, ...]
                predicted_skeleton: Tensor = out[:, [-2], ...]

                embedding: Tensor = vector_to_embedding(vector_scale, vector)
                out: Tensor = baked_embed_to_prob(embedding, baked, sigma(e))

                _loss_embed = loss_embed(
                    out, masks.gt(0).float()
                )  # out = [B, 2/3, X, Y, :w
                # Z?]
                _loss_prob = loss_prob(probability_map, masks.gt(0).float())
                _loss_skeleton = loss_skele(
                    predicted_skeleton, skele_masks.gt(0).float()
                )  # + skel_crossover_loss(predicted_skeleton, skele_masks.gt(0).float())

                logging.debug(f'\t{_loss_embed=}, {_loss_prob=}, {_loss_skeleton=}')

                # fuck this small amount of code.
                loss = (
                        (
                                cfg.TRAIN.LOSS_EMBED_RELATIVE_WEIGHT
                                * (1 if e > cfg.TRAIN.LOSS_EMBED_START_EPOCH else 0)
                                * _loss_embed
                        )
                        + (
                                cfg.TRAIN.LOSS_PROBABILITY_RELATIVE_WEIGHT
                                * (1 if e > cfg.TRAIN.LOSS_PROBABILITY_START_EPOCH else 0)
                                * _loss_prob
                        )
                        + (
                                cfg.TRAIN.LOSS_SKELETON_RELATIVE_WEIGHT
                                * (1 if e > cfg.TRAIN.LOSS_SKELETON_START_EPOCH else 0)
                                * _loss_skeleton
                        )
                )

                if torch.isnan(loss):
                    print(
                        f"Found NaN value in loss.\n\tLoss Embed: {_loss_embed}\n\tLoss Probability: {_loss_prob}"
                    )
                    print(f"\t{torch.any(torch.isnan(vector))}")
                    print(f"\t{torch.any(torch.isnan(embedding))}")
                    continue

            scaler.scale(loss).backward()
            scaler.step(optimizer)
            scaler.update()

            if e > swa_start:
                swa_model.update_parameters(model)

            _loss.append(loss.item())
            _embed.append(_loss_embed.item())
            _prob.append(_loss_prob.item())
            _skele.append(_loss_skeleton.item())

        avg_epoch_loss.append(mean(_loss))
        avg_epoch_embed_loss.append(mean(_embed))
        avg_epoch_prob_loss.append(mean(_prob))
        avg_epoch_skele_loss.append(mean(_skele))
        scheduler.step()

        if writer and (rank == 0):
            logging.info('writing to tensorboard')
            writer.add_scalar("lr", scheduler.get_last_lr()[-1], e)
            writer.add_scalar("Loss/train", avg_epoch_loss[-1], e)
            writer.add_scalar("Loss/embed", avg_epoch_embed_loss[-1], e)
            writer.add_scalar("Loss/prob", avg_epoch_prob_loss[-1], e)
            writer.add_scalar("Loss/skele-mask", avg_epoch_skele_loss[-1], e)
            write_progress(
                writer=writer,
                tag="Train",
                epoch=e,
                images=images,
                masks=masks,
                probability_map=probability_map,
                vector=vector,
                out=out,
                skeleton=skeleton,
                predicted_skeleton=predicted_skeleton,
                gt_skeleton=skele_masks,
            )

        # # Validation Step
        if e % 10 == 0 and valdiation_dataloader:
            logging.info('starting validation step')
            _loss, _embed, _prob, _skele = [], [], [], []
            for images, masks, skeleton, skele_masks, baked in valdiation_dataloader:
                with autocast(enabled=cfg.TRAIN.MIXED_PRECISION):  # Saves Memory!
                    with torch.no_grad():
                        out: Tensor = model(images)

                        probability_map: Tensor = out[:, [-1], ...]
                        predicted_skeleton: Tensor = out[:, [-2], ...]
                        vector: Tensor = out[:, 0:3:1, ...]

                        embedding: Tensor = vector_to_embedding(vector_scale, vector)
                        out: Tensor = baked_embed_to_prob(embedding, baked, sigma(e))

                        _loss_embed = loss_embed(out, masks.gt(0).float())
                        _loss_prob = loss_prob(probability_map, masks.gt(0).float())
                        _loss_skeleton = loss_prob(
                            predicted_skeleton, skele_masks.gt(0).float()
                        )

                        loss = (2 * _loss_embed) + (2 * _loss_prob) + _loss_skeleton

                        if torch.isnan(loss):
                            print(
                                f"Found NaN value in loss.\n\tLoss Embed: {_loss_embed}\n\tLoss Probability: {_loss_prob}"
                            )
                            print(f"\t{torch.any(torch.isnan(vector))}")
                            print(f"\t{torch.any(torch.isnan(embedding))}")
                            continue

                scaler.scale(loss)
                _loss.append(loss.item())
                _embed.append(_loss_embed.item())
                _prob.append(_loss_prob.item())
                _skele.append(_loss_skeleton.item())

            avg_val_loss.append(mean(_loss))
            avg_val_embed_loss.append(mean(_embed))
            avg_val_prob_loss.append(mean(_prob))
            avg_val_skele_loss.append(mean(_skele))

            if writer and (rank == 0):
                logging.info('writing validation to tensorboard')
                writer.add_scalar("Validation/train", avg_val_loss[-1], e)
                writer.add_scalar("Validation/embed", avg_val_embed_loss[-1], e)
                writer.add_scalar("Validation/prob", avg_val_prob_loss[-1], e)
                write_progress(
                    writer=writer,
                    tag="Validation",
                    epoch=e,
                    images=images,
                    masks=masks,
                    probability_map=probability_map,
                    vector=vector,
                    out=out,
                    skeleton=skeleton,
                    predicted_skeleton=predicted_skeleton,
                    gt_skeleton=skele_masks,
                )

        if rank == 0:
            epoch_range.desc = (
                    f"lr={scheduler.get_last_lr()[-1]:.3e}, Loss (train | val): "
                    + f"{avg_epoch_loss[-1]:.5f} | {avg_val_loss[-1]:.5f}"
            )

        logging.info('creating model state dict')
        state_dict = (
            model.module.state_dict()
            if hasattr(model, "module")
            else model.state_dict()
        )
        if e % 100 == 0:
            logging.info('writing intermediary state')
            torch.save(state_dict, cfg.TRAIN.SAVE_PATH + f"/test_{e}.trch")

    if rank == 0:
        state_dict = (
            model.module.state_dict()
            if hasattr(model, "module")
            else model.state_dict()
        )
        constants = {
            "cfg": cfg,
            "model_state_dict": state_dict,
            "optimizer_state_dict": optimizer.state_dict(),
            "avg_epoch_loss": avg_epoch_loss,
            "avg_epoch_embed_loss": avg_epoch_embed_loss,
            "avg_epoch_prob_loss": avg_epoch_prob_loss,
            "avg_epoch_skele_loss": avg_epoch_skele_loss,
            "avg_val_loss": avg_epoch_loss,
            "avg_val_embed_loss": avg_epoch_embed_loss,
            "avg_val_prob_loss": avg_epoch_prob_loss,
            "avg_val_skele_loss": avg_epoch_skele_loss,
        }
        try:
            torch.save(
                constants,
                f"{cfg.TRAIN.SAVE_PATH}/{os.path.split(writer.log_dir)[-1]}.trch",
            )
        except:
            print(
                f"Could not save at: {cfg.TRAIN.SAVE_PATH}/{os.path.split(writer.log_dir)[-1]}.trch"
                f"Saving at {os.getcwd()}/{os.path.split(writer.log_dir)[-1]}.trch instead"
            )

            torch.save(
                constants,
                f"{os.getcwd()}/{os.path.split(writer.log_dir)[-1]}.trch",
            )



import os.path
from typing import Tuple, Dict

import torch
import torch.nn.functional as F
from skimage.morphology import skeletonize
from torch import Tensor
from tqdm import tqdm
import kimimaro
import numpy as np
import skimage.io as io
import glob


def save_train_test_split(
    mask: Tensor, skeleton: Dict[int, Tensor], z_split: int, base: str
):
    """
    Splits a volume of binary masks and skeletons. You CANNOT naively just split the mask in two as skeletons of objects
    on the border might not be properly calculated.

    Saves pickled Dict[int, Tensor] to base+'_train.skeletons.trch' and base+'_validate.skeletons.trch'

    :param mask: Instance masks
    :param skeleton: Dict of skeleton of EVERY object in mask
    :param z_split: Z index of the train test split
    :param base: base filepath by which to save.

    :return: None
    """

    # train
    _mask = mask[..., 0 : z_split + 1 :]

    # assert 486 in _mask.unique()

    _skel = {}
    for u in _mask.unique():
        u = int(u)
        if u == 486:
            print("We got em...")
        if u == 0:
            continue
        if u in skeleton:
            _skel[u] = skeleton[u]
        # else:
        # print(f'Not in Train: {u}')

    torch.save(_skel, base + "_train.skeletons.trch")

    _mask = mask[..., z_split::]
    _skel = {}
    for u in _mask.unique():
        u = int(u)
        if u == 0:
            continue
        if u in skeleton:
            x = skeleton[u]
            x[:, 2] -= 150
            _skel[u] = x

    torch.save(_skel, base + "_validate.skeletons.trch")


def calculate_skeletons(mask: Tensor, scale: Tensor) -> Dict[int, Tensor]:
    """
    Calculates the skeleton of each object in mask

    :param mask: [C, X, Y, Z]
    :return: Dict[int, Tensor] dict of masks where int is the object id and Tensor is [3, K] skeletons
    """
    # raise DeprecationWarning('This function is slow and should no longer be used...')
    unique = torch.unique(mask)

    x, y, z = mask.shape

    if scale.sum() != 3:
        print(scale, x, y, z)

        large_mask = (
            F.interpolate(
                mask.unsqueeze(0).unsqueeze(0).float(),
                size=torch.tensor([x, y, z]).mul(scale).float().round().int().tolist(),
                mode="nearest",
            )
            .squeeze()
            # .cuda()
            .int()
        )
    else:
        large_mask = mask.clone()

    large_mask_unique = large_mask.unique().tolist()
    unique_list = unique.tolist()

    for u in unique:
        assert u in large_mask_unique, 'Downscaled too much!'

    for u in large_mask_unique:
        assert u in unique_list, 'Downscaled too much!'

    # assert torch.allclose(
    #     unique.cuda(), torch.unique(large_mask)
    # ), f"{unique=}, {large_mask.unique()=}"

    # large_mask = mask

    unique, counts = torch.unique(large_mask, return_counts=True)

    output = {}

    for id in tqdm(unique):
        if id == 0:
            continue

        temp = large_mask == id
        nonzero = torch.nonzero(temp)

        lower = nonzero.min(0)[0]
        upper = nonzero.max(0)[0]

        # print(f'{id.item()=}, {lower=}, {upper=}')

        upper[upper.sub(lower) == 0] += 1

        # Get just the region of a particular instance of the binary image...
        temp = temp[
            lower[0].item() : upper[0].item(),  # x
            lower[1].item() : upper[1].item(),  # y
            lower[2].item() : upper[2].item(),  # z
        ].float()

        _x = upper[0] - lower[0]
        _y = upper[1] - lower[1]
        _z = upper[2] - lower[2]

        # Calculate the binary skeleton of that image...
        skeleton = skeletonize(temp.cpu().numpy(), method="lee")
        skeleton = torch.from_numpy(skeleton).unsqueeze(0).unsqueeze(0)

        offset = lower.cpu().div(scale)  # , rounding_mode='trunc')
        # offset = offset

        if torch.nonzero(skeleton).shape[0] != 0:
            skel = torch.nonzero(skeleton.squeeze(0).squeeze(0)).div(scale).add(offset)
            # print(skel)
            output[int(id)] = skel
        else:
            _nonzoer = torch.nonzero(temp.cpu()).float()
            _nonzoer = _nonzoer.unsqueeze(0) if _nonzoer.ndim == 1 else _nonzoer
            output[int(id)] = _nonzoer.mean(0).div(scale).add(offset).unsqueeze(0)

        assert (
            output[int(id)].shape[0] > 0 and output[int(id)].ndim > 1
        ), f"{temp.nonzero().shape=}, {lower=} {id}, {output[int(id)].shape}"

    return output


def _calculate_skeletons(mask: Tensor, scale: Tensor) -> Dict[int, Tensor]:
    """

    image of shape [X, Y, Z,] with int masks
    returns dict[int, Tensor]

    :param mask:
    :param scale:
    :return:
    """

    skels = kimimaro.skeletonize(
        mask.numpy(),
        anisotropy=scale,
        progress=True,
        dust_threshold=0,
        fill_holes=False,
        fix_avocados=False,
        parallel=0
    )
    print('test1234')
    skels = {k:torch.from_numpy(v.vertices).div(torch.tensor(scale)).round().long() for k, v in skels.items()}


    return skels


def create_gt_skeletons(base_dir, mask_filter, scale: Tuple[float, float, float]):
    files = glob.glob(os.path.join(base_dir, f"*{mask_filter}.tif"))
    files = [b[:-11:] for b in files]

    scale = torch.tensor(scale)

    for f in files:
        print('attempting to skeletonize ', f+f"{mask_filter}.tif" )
        mask = io.imread(f + f"{mask_filter}.tif")
        mask = torch.from_numpy(mask.astype(np.int32))
        mask = mask.permute((1, 2, 0))

        output = calculate_skeletons(mask, scale)

        for u in mask.unique():
            if u == 0:
                continue
            assert int(u) in output, f"{f}, {u}, {output.keys()=}"

        torch.save(output, f + ".skeletons.trch")
        print("SAVED", f + ".skeletons.trch")


if __name__ == "__main__":
    import skimage.io as io
    import numpy as np
    import glob

    """
    The move -> Calculate the skeleton of each instance and save it as a tensor of nonzero indicies
    For each pixel in the instance mask, we now need to calculate which skeleton point we need to point to
    To do this, we take the index of each point, find the skeleton point closest to it, andreplace it
    """

    image = io.imread('/home/chris/Dropbox (Partners HealthCare)/skoots-experiments/data/mitochondria/train/external/kasthuri.labels.tif')
    image = torch.from_numpy(image.transpose((1, 2, 0)))
    skels = calculate_skeletons(image, (1, 1, 7))

    skeletons = torch.zeros_like(image)
    print(image.shape, skeletons.shape)

    for id, skel in skels.items():
        print(skel.shape)
        skeletons[skel[:,0].long(), skel[:,1].long(), skel[:,2].div(7).round().long()] = 1

    io.imsave('/home/chris/Desktop/skeletons_test.tif', skeletons.permute((2, 0 ,1).numpy()))

    # bases = [
    #     # '/home/chris/Dropbox (Partners HealthCare)/trainMitochondriaSegmentation/data/train/',
    #     "/home/chris/Dropbox (Partners HealthCare)/trainMitochondriaSegmentation/data/toBeSkeletonized/",
    # ]
    # if not torch.cuda.is_available():
    #     raise RuntimeError("NO CUDA")
    #
    # # Z_SCALE = 2
    # SCALE = torch.tensor([0.2, 0.2, 3])
    # # SCALE = torch.tensor([1,1,1])
    #
    # bases = glob.glob(bases[0] + "*.labels.tif")
    # bases = [b[:-11:] for b in bases]
    #
    # for base in bases:
    #     mask = io.imread(base + ".labels.tif")
    #     mask = torch.from_numpy(mask.astype(np.int32))
    #
    #     mask = mask.permute((1, 2, 0))
    #     x, y, z = mask.shape
    #
    #     output = calculate_skeletons(mask, SCALE)
    #
    #     torch.save(output, base + ".skeletons.trch")
    #     print("SAVED", base + ".skeletons.trch")
    #
    #     save_train_test_split(mask, output, 150, base)



import argparse
import os.path

import torch
import torch.multiprocessing as mp
import torch.nn as nn

from skoots.config import get_cfg_defaults
from skoots.lib.mp_utils import find_free_port
from skoots.lib.utils import cfg_to_bism_model
from skoots.train.engine import train

if __name__ == '__main__':

    torch.set_float32_matmul_precision("high")

    cfg = get_cfg_defaults()
    file = '/home/chris/Dropbox (Partners HealthCare)/skoots-experiments/configs/mitochondria/train_skoots_on_mitochondria.yaml'
    if os.path.exists(file):
        cfg.merge_from_file(file)
    cfg.freeze()

    model: nn.Module = cfg_to_bism_model(cfg)  # This is our skoots torch model

    if cfg.TRAIN.PRETRAINED_MODEL_PATH:
        checkpoint = torch.load(cfg.TRAIN.PRETRAINED_MODEL_PATH[0])
        state_dict = (
            checkpoint
            if not "model_state_dict" in checkpoint
            else checkpoint["model_state_dict"]
        )
        model.load_state_dict(state_dict)

    port = find_free_port()
    world_size = cfg.SYSTEM.NUM_GPUS if cfg.TRAIN.DISTRIBUTED else 1

    mp.spawn(train, args=(port, world_size, model, cfg), nprocs=world_size, join=True)
